"""
This module contains functions for splitting a torch dataset into parts.
"""
from typing import List, Tuple, Generator, Optional, Dict

import numpy as np
from torch.utils.data import Dataset, Subset
from tqdm import tqdm


def train_test_split(dataset: Dataset, p: float = 0.2) -> Tuple[Subset, Subset]:
    """
    Splits the data into two parts, keeping the proportions of the classes.

    Args:
        dataset: Torch dataset object.
        p: Proportion of the test sample, must be from 0 to 1.

    Returns:
        A tuple ``(train_ds, test_ds)``, where train_ds and test_ds
            are subsets.
    """
    n = int(1./p)
    return next(k_fold(dataset=dataset, n=n))


def k_fold(dataset: Dataset, n: int) -> Generator[Tuple[Subset, Subset], None, None]:
    """
    K-Folds cross-validator.

    Args:
        dataset: Torch dataset object.
        n: Number of folds.

    Yields:
        A tuple ``(train_ds, test_ds)``, where train_ds and test_ds
            are subsets.
    """
    fold_indices = split_data(dataset, n)
    for i in range(n):
        test_indices = fold_indices[i]
        test_ds = Subset(dataset, test_indices)
        train_indices = np.concatenate([fold_indices[j] for j in range(n) if j != i])
        train_ds = Subset(dataset, train_indices)
        yield train_ds, test_ds

def split_data(dataset: Dataset, n: int, verbose: bool = False) -> List[np.ndarray]:
    """
    Splits the data into n parts, keeping the proportions of the classes.

    Args:
        dataset: Torch dataset object.
        n: Number of parts.
        verbose: If `True` prints information about splitting.

    Returns:
        A list of indices for each part.
    """
    classes_of_imgs = _extract_classes(dataset)
    classes = np.unique(classes_of_imgs)
    fold_indices = [[] for _ in range(n)]
    for cl in classes:
        indices_of_class = np.random.permutation(np.nonzero(classes_of_imgs==cl)[0])
        if verbose:
            print(f"Class {cl} contains {indices_of_class.size} samples.")
        lengths = [int(indices_of_class.size / n) for _ in range(n)]
        remainder = indices_of_class.size - sum(lengths)
        # add 1 to all the lengths in round-robin fashion until the remainder is 0
        for i in range(remainder):
            lengths[i % n] += 1
        start = 0
        if verbose:
            print(' '.join(map(str, lengths)))
        for i, length in enumerate(lengths):
            fold_indices[i].append(indices_of_class[start : start + length])
            start += length
    return [np.concatenate(fold) for fold in fold_indices]


def undersampling(dataset: Dataset, n: Optional[int] = None, verbose: bool = False) -> Subset:
    """
    A method for balancing uneven datasets by keeping all data in the
    minority class and decreasing the size of the majority class.

    Args:
        dataset: Torch dataset object.
        n: Number of samples in each class.
        verbose: If `True` prints information about undersampling.

    Returns:
        Balanced subset.
    """
    classes_of_imgs = _extract_classes(dataset)
    classes = np.unique(classes_of_imgs)
    indices_of_classes = []
    min_size = len(classes_of_imgs)
    for cl in classes:
        indices_of_class = np.random.permutation(np.nonzero(classes_of_imgs==cl)[0])
        indices_of_classes.append(indices_of_class)
        if verbose:
            print(f"Class {cl} contains {indices_of_class.size} samples.")
        if indices_of_class.size < min_size:
            min_size = indices_of_class.size
    min_size = min_size if n is None else n
    print(f"New size of any class {min_size} samples.")
    indices = np.concatenate([x[:min_size] for x in indices_of_classes])
    return Subset(dataset, indices)


def _extract_classes(dataset: Dataset) -> np.ndarray:
    """
    Returns the class for each sample.

    Args:
        dataset: Torch dataset object.
    """
    classes_of_imgs = []
    for i in tqdm(range(len(dataset)), desc='prepare dataset'):
        img, target = dataset.__getitem__(i)
        classes_of_imgs.append(target)
    return np.array(classes_of_imgs)


def dataset_info(dataset: Dataset, verbose: bool = False) -> Dict[int, int]:
    """
    Returns number of samples in each class

    Args:
        dataset: Torch dataset object.
        verbose: If `True` prints information about classes.

    Returns:
        Dictionary `{class_id: number_of_samples}`.
    """
    classes_of_imgs = _extract_classes(dataset)
    classes = np.unique(classes_of_imgs)
    class_samples = {}
    for cl in classes:
        indices_of_class = np.nonzero(classes_of_imgs==cl)[0]
        class_samples[cl] = indices_of_class.size
        if verbose:
            print(f"Class {cl} contains {indices_of_class.size} samples.")
    return class_samples


def get_dataset_mean_std(dataset: Dataset) -> Tuple[Tuple, Tuple]:
    """
    Compute mean and std of dataset.

    Args:
        dataset: Torch dataset object.

    Returns:
          Tuple(mean, std)
    """
    shape = dataset.__getitem__(0)[0].shape
    one_channel = len(shape) == 2

    if one_channel:
        psum = np.zeros(1)
        psum_sq = np.zeros(1)
        pixels = np.zeros(1)
        for i in tqdm(range(len(dataset)), desc='computing mean and std'):
            img, target = dataset.__getitem__(i)
            psum += img.sum().numpy()
            psum_sq += (img ** 2).sum().numpy()
            pixels += img.numel()
    else:
        psum = np.zeros(shape[0])
        psum_sq = np.zeros(shape[0])
        pixels = np.zeros(1)
        for i in tqdm(range(len(dataset)), desc='computing mean and std'):
            img, target = dataset.__getitem__(i)
            psum += img.sum(dim=[1, 2]).numpy()
            psum_sq += (img ** 2).sum(dim=[1, 2]).numpy()
            c, w, h = img.size()
            pixels += w * h
    mean = psum / pixels
    var = (psum_sq / pixels) - (mean ** 2)
    std = np.sqrt(var)
    return tuple(mean), tuple(std)



from abc import ABC, abstractmethod
from datetime import datetime as dt
from math import factorial

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt


class TimeSeriesGenerator:

    def __init__(self, params: dict):
        self.ts_type = params.get('ts_type', 'smooth_normal')
        self.ts_length = params.get('length', 1000)
        self.seed = params.get('seed', self.__define_seed())
        self.__define_randomness()

        self.ts_types = {'sin': SinWave,
                         'random_walk': RandomWalk,
                         'auto_regression': AutoRegression,
                         'smooth_normal': SmoothNormal}
        self.params = params

    def __define_seed(self):
        self.seed = dt.now().second

    def __define_randomness(self):
        np.random.seed(self.seed)

    def get_ts(self):
        if self.ts_type not in self.ts_types.keys():
            raise ValueError('ts_type must be one of the following: "sin", "smooth_normal", "random_walk", '
                             '"auto_regression"')
        else:
            ts_class = self.ts_types[self.ts_type](self.params)
            return ts_class.get_ts()


class DefaultTimeSeries(ABC):
    def __init__(self, params: dict):
        self.ts_length = params.get('length', 1000)

    @abstractmethod
    def get_ts(self):
        NotImplementedError()


class SinWave(DefaultTimeSeries):
    def __init__(self, params: dict):
        super().__init__(params)
        self.amplitude = params.get('amplitude', 10)
        self.period = params.get('period', self.ts_length)

    def get_ts(self):
        time_index = np.arange(0, self.ts_length)
        sine_wave = self.amplitude * np.sin(2 * np.pi / self.period * time_index)
        return sine_wave


class RandomWalk(DefaultTimeSeries):
    def __init__(self, params: dict):
        super().__init__(params)
        self.start_val = params.get('start_val', 36.6)

    def get_ts(self):
        time_index = pd.Series(np.arange(0, self.ts_length))
        random_walk = pd.Series(np.cumsum(np.random.randn(self.ts_length)) + self.start_val, index=time_index)
        return np.array(random_walk)


class AutoRegression(DefaultTimeSeries):
    def __init__(self, params: dict):
        super().__init__(params)
        self.ar_params = params.get('ar_params', [0.5, -0.3, 0.2])
        self.initial_values = params.get('initial_values', None)

    def get_ts(self):
        time_index = pd.Series(np.arange(0, self.ts_length))
        ar_process = np.zeros(self.ts_length)

        for i in range(len(ar_process)):
            if i < len(self.ar_params):
                if self.initial_values is None:
                    ar_process[i] = np.random.normal(0, 1, 1)
                else:
                    ar_process[i] = self.initial_values[i]
            else:
                ar_process[i] = np.sum(self.ar_params * ar_process[i - len(self.ar_params):i].ravel(),
                                       axis=0) + np.random.normal(0, 1)
        # return np.array(pd.Series(ar_process[:, 0], index=time_index))
        return ar_process


class SmoothNormal(DefaultTimeSeries):
    def __init__(self, params: dict):
        super().__init__(params)
        self.window_size = params.get('window_size', int(self.ts_length / 3))
        self.__check_window_size()

    def __check_window_size(self):
        # window size must be odd for smoothing ts with Savitzky-Golay filter
        if self.window_size % 2 == 0:
            self.window_size += 1

    def get_ts(self):
        normal_ts = np.random.normal(0, 1, self.ts_length) + 100
        return self.savitzky_golay(y=normal_ts, window_size=self.window_size, order=3)

    @staticmethod
    def savitzky_golay(y: np.array, window_size: int, order: int = 3, deriv: int = 0, rate: int = 1):
        """
        Smooth (and optionally differentiate) data with a Savitzky-Golay filter.

        Args:
            y: array_like, shape (N,) the values of the time history of the signal.
            window_size: int (odd) the length of the window. Must be an odd integer number.
            order: int (optional) the order of the polynomial used in the filtering.
            deriv: int (optional) the order of the derivative to compute.
            rate: int or float (optional) sampling rate of y. Default is 1.

        Returns:

        Notes:
            Taken from https://scipy.github.io/old-wiki/pages/Cookbook/SavitzkyGolay
        """
        try:
            window_size = np.abs(int(window_size))
            order = np.abs(order)
        except ValueError:
            raise ValueError("window_size and order have to be of type int")
        if window_size % 2 != 1 or window_size < 1:
            raise TypeError("window_size size must be a positive odd number")
        if window_size < order + 2:
            raise TypeError("window_size is too small for the polynomials order")
        order_range = range(order + 1)
        half_window = (window_size - 1) // 2
        # precompute coefficients
        b = np.mat([[k ** i for i in order_range] for k in range(-half_window, half_window + 1)])
        m = np.linalg.pinv(b).A[deriv] * rate ** deriv * factorial(deriv)
        # pad the signal at the extremes with
        # values taken from the signal itself
        firstvals = y[0] - np.abs(y[1:half_window + 1][::-1] - y[0])
        lastvals = y[-1] + np.abs(y[-half_window - 1:-1][::-1] - y[-1])
        y = np.concatenate((firstvals, y, lastvals))
        return np.convolve(m[::-1], y, mode='valid')


if __name__ == '__main__':
    ts_config = {
        'ts_type': 'random_walk',
        # 'ts_type': 'sin',
        'length': 1000,

        # for sin wave
        'amplitude': 10,
        'period': 500,

        # for random walk
        'start_val': 36.6,

        # for auto regression
        'ar_params': [0.5, -0.3, 0.2],
        'initial_values': None,

        # for smooth normal
        'window_size': 300}

    ts_generator = TimeSeriesGenerator(ts_config)
    ts = ts_generator.get_ts()

    plt.plot(ts)
    plt.show()
    _ = 1



import logging
from typing import Optional

import numpy as np
import pandas as pd
from fedot.api.main import Fedot
from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from sklearn.decomposition import PCA

from fedot_ind.core.metrics.evaluation import PerformanceAnalyzer


class TimeSeriesRegression:

    def __init__(self, params: Optional[OperationParameters] = None):
        self.strategy = params.get('strategy', 'statistical')
        self.model_hyperparams = params.get('model_params')
        self.generator_runner = params.get('generator_class')
        self.dataset_name = params.get('dataset')
        self.output_folder = params.get('output_folder', None)
        self.use_cache = params.get('use_cache', False)

        self.logger = logging.getLogger('TimeSeriesClassifier')
        self.pca = PCA(n_components=params.get('explained_variance', 0.9), svd_solver='full')

        self.logger.info('TimeSeriesRegression solver initialised')

    def __check_multivariate_data(self, data: pd.DataFrame) -> bool:
        """Method for checking if the data is multivariate.

        Args:
            X: pandas DataFrame with features

        Returns:
            True if data is multivariate, False otherwise

        """
        if isinstance(data.iloc[0, 0], pd.Series):
            return True
        else:
            return False

    def _init_input_data(self, X: pd.DataFrame, target: np.ndarray) -> InputData:
        """Method for initialization of InputData object from pandas DataFrame and numpy array with target values.

        Args:
            X: pandas DataFrame with features
            y: numpy array with target values

        Returns:
            InputData object convenient for FEDOT framework

        """
        y = np.array([float(i) for i in target])

        is_multivariate_data = self.__check_multivariate_data(X)
        if is_multivariate_data:
            input_data = InputData(idx=np.arange(len(X)),
                                   features=np.array(X.values.tolist()),
                                   target=y.reshape(-1, 1),
                                   task=Task(TaskTypesEnum.regression),
                                   data_type=DataTypesEnum.image)
        else:
            input_data = InputData(idx=np.arange(len(X)),
                                   features=X.values,
                                   target=np.ravel(y).reshape(-1, 1),
                                   task=Task(TaskTypesEnum.regression),
                                   data_type=DataTypesEnum.table)

        return input_data

    def fit_regressor(self, features, target):
        self.logger.info('Start fitting regressor')
        self.predictor = Fedot(**self.model_hyperparams)
        self.predictor.fit(features, target)
        return self.predictor

    def fit(self, features,
            target: np.ndarray = None,
            **kwargs) -> object:
        """
        Method for fitting pipeline on train data. It also tunes pipeline and updates it with categorical features.

        Args:
            features: pandas DataFrame with features
            target: numpy array with target values

        Returns:
            fitted FEDOT model as object of ``Pipeline`` class

        """

        self.logger.info('Start fitting pipeline')
        self.train_target = np.array([float(i) for i in target])
        self.train_input_data = self._init_input_data(features, target)
        extracted_train_features = self.generator_runner.transform(input_data=self.train_input_data,
                                                         use_cache=self.use_cache)
        train_size = extracted_train_features.features.shape
        self.train_features = extracted_train_features.features.reshape(train_size[0], train_size[1] * train_size[2])

        self.logger.info('Start applying PCA')
        self.pca_train_features = self.pca.fit_transform(self.train_features)
        self.logger.info(f'PCA n_components left: {self.pca.n_components_}')

        self.fit_regressor(self.pca_train_features, self.train_target)

        return self.predictor

    def predict(self, features: pd.DataFrame, target: np.array):
        """Method for prediction on test data.

        Args:
            features: pandas DataFrame with features
            target: numpy array with target values

        Returns:
            dict with predicted values and probabilities

        """
        self.logger.info('Start prediction')
        self.test_target = np.array([float(i) for i in target])
        self.test_input_data = self._init_input_data(features, target)
        extracted_test_features = self.generator_runner.transform(input_data=self.test_input_data,
                                                             use_cache=self.use_cache)
        test_size = extracted_test_features.features.shape
        self.test_features = extracted_test_features.features.reshape(test_size[0], test_size[1] * test_size[2])

        self.pca_test_features = self.pca.transform(self.test_features)
        self.predicted_labels = self.predictor.predict(features=self.pca_test_features)

        return self.predicted_labels

    def get_metrics(self, target, metric_names: Optional[list] = None):
        analyzer = PerformanceAnalyzer()
        return analyzer.calculate_metrics(target=self.test_target,
                                          predicted_labels=self.predicted_labels,
                                          predicted_probs=None,
                                          target_metrics=metric_names)



"""
This module contains classes for CNN structure optimization.
"""

import logging
import os
from typing import Callable, Dict, List, Optional, Type
from abc import ABC, abstractmethod
from functools import partial

import torch
from torchvision.models import ResNet

from fedot_ind.core.architecture.abstraction.writers import WriterComposer, TFWriter, CSVWriter, Writer
from fedot_ind.core.architecture.experiment.nn_experimenter import NNExperimenter, FitParameters
from fedot_ind.core.metrics.loss.svd_loss import OrthogonalLoss, HoyerLoss
from fedot_ind.core.operation.decomposition.decomposed_conv import DecomposedConv2d
from fedot_ind.core.operation.optimization.sfp_tools import percentage_filter_zeroing, energy_filter_zeroing, prune_resnet, load_sfp_resnet_model
from fedot_ind.core.operation.optimization.svd_tools import energy_svd_pruning, decompose_module, load_svd_state_dict


class StructureOptimization(ABC):
    """Generalized class for model structure optimization."""

    def __init__(
            self,
            description,
    ) -> None:
        self.logger = logging.getLogger(self.__class__.__name__)
        self.description = description

    @abstractmethod
    def fit(
            self,
            exp: NNExperimenter,
            params: FitParameters,
            ft_params: Optional[FitParameters] = None
    ) -> None:
        """Run model training with optimization.

        Args:
            exp: An instance of the experimenter class, e.g. ``ClassificationExperimenter``.
            params: An object containing training parameters.
            ft_params: An object containing fine-tuning parameters for optimized model.
        """
        raise NotImplementedError

    @abstractmethod
    def load_model(self, exp: NNExperimenter, state_dict_path: str) -> None:
        """Loads the optimized model into the experimenter.

        Args:
            exp: An instance of the experimenter class, e.g. ``ClassificationExperimenter``.
            state_dict_path: Path to state_dict file.
        """
        raise NotImplementedError


class SVDOptimization(StructureOptimization):
    """Singular value decomposition for model structure optimization.

    Args:
        energy_thresholds: List of pruning hyperparameters.
        decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
        forward_mode: ``'one_layer'``, ``'two_layers'`` or ``'three_layers'`` forward pass calculation method.
        hoer_loss_factor: The hyperparameter by which the Hoyer loss function is
            multiplied.
        orthogonal_loss_factor: The hyperparameter by which the orthogonal loss
            function is multiplied.
    """

    def __init__(
            self,
            energy_thresholds: List[float] = [0.9, 0.95, 0.99, 0.999],
            decomposing_mode: str = 'channel',
            forward_mode: str = 'one_layer',
            hoer_loss_factor: float = 0.1,
            orthogonal_loss_factor: float = 10,
    ) -> None:
        super().__init__(
            description=(
                f"_SVD_{decomposing_mode}_O-{orthogonal_loss_factor}"
                f"_H-{hoer_loss_factor}"
            )
        )
        self.energy_thresholds = energy_thresholds
        self.decomposing_mode = decomposing_mode
        self.forward_mode = forward_mode
        self.hoer_loss = HoyerLoss(hoer_loss_factor)
        self.orthogonal_loss = OrthogonalLoss(orthogonal_loss_factor)
        self.finetuning = False

    def fit(
            self,
            exp: NNExperimenter,
            params: FitParameters,
            ft_params: Optional[FitParameters] = None
    ) -> None:
        """Run model training with optimization.

        Args:
            exp: An instance of the experimenter class, e.g. ``ClassificationExperimenter``.
            params: An object containing training parameters.
            ft_params: An object containing fine-tuning parameters for optimized model.
        """
        exp.name += self.description
        self.finetuning = False
        writer = CSVWriter(os.path.join(params.summary_path, params.dataset_name, exp.name, params.description))
        size = {'size': exp.size_of_model(), 'params': exp.number_of_model_params()}
        writer.write_scores(phase='size', scores=size, x='default')
        self.logger.info(f"Default size: {size['size']:.2f} Mb")
        decompose_module(exp.model, self.decomposing_mode, forward_mode=self.forward_mode)
        exp.model.to(exp.device)
        size = {'size': exp.size_of_model(), 'params': exp.number_of_model_params()}
        writer.write_scores(phase='size', scores=size, x='decomposed')
        self.logger.info(f"SVD decomposed size: {size['size']:.2f} Mb")

        exp.fit(p=params, model_losses=self.__loss)
        self.optimize(exp=exp, params=params, ft_params=ft_params, writer=writer)

    def optimize(
            self,
            exp: NNExperimenter,
            params: FitParameters,
            ft_params: Optional[FitParameters],
            writer: Writer
    ) -> None:
        """Prunes the trained model at the given thresholds.

        Args:
            exp: An instance of the experimenter class, e.g. ``ClassificationExperimenter``.
            params: An object containing training parameters.
            ft_params: An object containing fine-tuning parameters for optimized model.
            writer: Object for recording metrics.
        """
        self.finetuning = True
        models_path = os.path.join(params.models_path, params.dataset_name, exp.name, params.description)
        exp.save_model(os.path.join(models_path, 'trained'), state_dict=False)
        for e in self.energy_thresholds:
            str_e = f'e_{e}'
            exp.load_model(os.path.join(models_path, 'trained'), state_dict=False)
            exp._apply_function(
                func=partial(energy_svd_pruning, energy_threshold=e),
                condition=lambda x: isinstance(x, DecomposedConv2d)
            )
            exp.save_model(os.path.join(models_path, str_e))
            scores = {'size': exp.size_of_model(), 'params': exp.number_of_model_params()}
            writer.write_scores(phase='size', scores=scores, x=str_e)
            scores.update(exp.val_loop(params.val_dl, params.class_metrics))
            writer.write_scores(phase='pruning', scores=scores, x=str_e)
            self.logger.info((f"pruning with {e=}, size: {scores['size']:.2f} Mb, "
                             f"{exp.metric}: {scores[exp.metric]:.4f}"))
            if ft_params is not None:
                exp.best_score = scores[exp.metric]
                exp.fit(
                    p=ft_params,
                    phase=str_e,
                    model_losses=self.__loss,
                    start_epoch=params.num_epochs,
                    initial_validation=True
                )

    def __loss(self, model: torch.nn.Module) -> Dict[str, torch.Tensor]:
        """
        Computes the orthogonal and the Hoer loss for the model.

        Args:
            model: CNN model with DecomposedConv layers.

        Returns:
            A dict ``{loss_name: loss_value}``
                where loss_name is a string and loss_value is a floating point tensor with a single value.
        """
        losses = {'orthogonal_loss': self.orthogonal_loss(model)}
        if not self.finetuning:
            losses['hoer_loss'] = self.hoer_loss(model)
        return losses

    def load_model(self, exp: NNExperimenter, state_dict_path: str) -> None:
        """Loads the optimized model into the experimenter.

        Args:
            exp: An instance of the experimenter class, e.g. ``ClassificationExperimenter``.
            state_dict_path: Path to state_dict file.
        """
        load_svd_state_dict(
            model=exp.model,
            state_dict_path=state_dict_path,
            decomposing_mode=self.decomposing_mode,
            forward_mode=self.forward_mode
        )
        exp.model.to(exp.device)
        self.logger.info("Model state dict loaded.")


class SFPOptimization(StructureOptimization):
    """Soft filter pruning for model structure optimization.

    Args:
        zeroing_fn: Partially initialized filter zeroing function.
        model_class: The class of models to which the final pruning function is applicable.
        final_pruning_fn: Function implementing the model final pruning of the ``model_class``.
        load_model_fn: Function implementing the model loading of the ``model_class``.
    """

    def __init__(
            self,
            zeroing_fn: partial = partial(percentage_filter_zeroing, pruning_ratio=0.2),
            model_class: Type = ResNet,
            final_pruning_fn: Callable = prune_resnet,
            load_model_fn: Callable = load_sfp_resnet_model
    ) -> None:
        description = f"_SFP"
        for k, v in zeroing_fn.keywords.items():
            description += f"_{k}-{v}"
        super().__init__(
            description=description,
        )
        self.zeroing_fn = zeroing_fn
        self.pruning_fn = final_pruning_fn
        self.model_class = model_class
        self.load_model_fn = load_model_fn

    def fit(
            self,
            exp: NNExperimenter,
            params: FitParameters,
            ft_params: Optional[FitParameters] = None
    ) -> None:
        """Run model training with optimization.

        Args:
            exp: An instance of the experimenter class, e.g. ``ClassificationExperimenter``.
            params: An object containing training parameters.
            ft_params: An object containing fine-tuning parameters for optimized model.
        """
        exp.name += self.description
        exp.fit(p=params, filter_pruning=dict(func=self.zeroing_fn, condition=lambda x: isinstance(x, torch.nn.Conv2d)))
        if isinstance(exp.model, self.model_class):
            self.final_pruning(exp=exp, params=params, ft_params=ft_params)
        else:
            self.logger.warning(f'Final pruning function "{self.pruning_fn.__name__}"' +
                                f'supports only {self.model_class.__name__} models.')

    def final_pruning(
            self,
            exp: NNExperimenter,
            params: FitParameters,
            ft_params: Optional[FitParameters],
    ) -> None:
        """Final model pruning after training.

        Args:
            exp: An instance of the experimenter class, e.g. ``ClassificationExperimenter``.
            params: An object containing training parameters.
            ft_params: An object containing fine-tuning parameters for optimized model.
        """
        writer = CSVWriter(os.path.join(params.summary_path, params.dataset_name, exp.name, params.description))
        size = {'size': exp.size_of_model(), 'params': exp.number_of_model_params()}
        writer.write_scores(phase='size', scores=size, x='default')
        self.logger.info(f"Default size: {size['size']:.2f} Mb")

        exp.model = self.pruning_fn(model=exp.model)
        exp.model.to(exp.device)
        exp.save_model(
            os.path.join(params.models_path, params.dataset_name, exp.name, params.description, 'pruned')
        )
        size = {'size': exp.size_of_model(), 'params': exp.number_of_model_params()}
        writer.write_scores(phase='size', scores=size, x='pruned')
        self.logger.info(f"Pruned size: {size['size']:.2f} Mb")

        exp.best_score = 0
        if ft_params is not None:
            exp.fit(p=ft_params, phase='pruned', start_epoch=params.num_epochs)

    def load_model(self, exp: NNExperimenter, state_dict_path: str) -> None:
        """Loads the optimized model into the experimenter.

        Args:
            exp: An instance of the experimenter class, e.g. ``ClassificationExperimenter``.
            state_dict_path: Path to state_dict file.
        """
        try:
            exp.model = self.load_model_fn(model=exp.model, state_dict_path=state_dict_path)
            self.logger.info("Model state dict loaded.")
        except Exception:
            self.logger.error(f'Loading function "{self.load_model_fn.__name__}"' +
                              f"can't load {self.model_class.__name__} model.")



"""This module contains functions for working with singular value decomposition.

Model decomposition, pruning by threshold, decomposed model loading.
"""
from typing import Optional, Callable

import torch
from torch.nn.modules import Module
from torch.nn.modules.conv import Conv2d

from fedot_ind.core.operation.decomposition.decomposed_conv import DecomposedConv2d


def energy_svd_pruning(conv: DecomposedConv2d, energy_threshold: float) -> None:
    """Prune the weight matrices to the energy_threshold (in-place).
    Args:
        conv: The optimizable layer.
        energy_threshold: pruning hyperparameter must be in the range (0, 1].
            the lower the threshold, the more singular values will be pruned.
    Raises:
        Assertion Error: If ``conv.decomposing`` is False.
        Assertion Error: If ``energy_threshold`` is not in (0, 1].
    """
    assert conv.decomposing is not None, "for pruning, the model must be decomposed"
    assert 0 < energy_threshold <= 1, "energy_threshold must be in the range (0, 1]"

    S, indices = conv.S.sort()
    U = conv.U[:, indices]
    Vh = conv.Vh[indices, :]
    sum = (S ** 2).sum()
    threshold = energy_threshold * sum
    for i, s in enumerate(S):
        sum -= s ** 2
        if sum < threshold:
            conv.set_U_S_Vh(U[:, i:].clone(), S[i:].clone(), Vh[i:, :].clone())
            break


def decompose_module(
        model: Module,
        decomposing_mode: Optional[str] = None,
        forward_mode: str = 'one_layer',
) -> None:
    """Replace Conv2d layers with DecomposedConv2d layers in module (in-place).

    Args:
        model: Decomposable module.
        decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
            If ``None`` replace layers without decomposition.
        forward_mode: ``'one_layer'``, ``'two_layers'`` or ``'three_layers'`` forward pass calculation method.
    """
    for name, module in model.named_children():
        if len(list(module.children())) > 0:
            decompose_module(module, decomposing_mode=decomposing_mode, forward_mode=forward_mode)

        if isinstance(module, Conv2d):
            new_module = DecomposedConv2d(
                base_conv=module,
                decomposing_mode=decomposing_mode,
                forward_mode=forward_mode
            )
            setattr(model, name, new_module)


def _load_svd_params(model, state_dict, prefix='') -> None:
    """Loads state_dict to DecomposedConv2d layers in model."""
    for name, module in model.named_children():
        if len(list(module.children())) > 0:
            _load_svd_params(module, state_dict, prefix=f'{prefix}{name}.')

        if isinstance(module, DecomposedConv2d):
            module.set_U_S_Vh(
                u=state_dict[f'{prefix}{name}.U'],
                s=state_dict[f'{prefix}{name}.S'],
                vh=state_dict[f'{prefix}{name}.Vh']
            )


def load_svd_state_dict(
        model: Module,
        decomposing_mode: str,
        state_dict_path: str,
        forward_mode: str = 'one_layer',
) -> None:
    """Loads SVD state_dict to model.

    Args:
        model: An instance of the base model.
        decomposing_mode: ``'channel'`` or ``'spatial'`` weights reshaping method.
        state_dict_path: Path to state_dict file.
        forward_mode: ``'one_layer'``, ``'two_layers'`` or ``'three_layers'`` forward pass calculation method.
    """
    state_dict = torch.load(state_dict_path, map_location='cpu')
    decompose_module(model=model, decomposing_mode=decomposing_mode, forward_mode=forward_mode)
    _load_svd_params(model, state_dict)
    model.load_state_dict(state_dict)



import logging
from pathlib import Path
from typing import List, Union, Dict
from typing import Optional

import matplotlib.patches as mpatches
import numpy as np
import pandas as pd
from fedot.api.main import Fedot
from fedot.core.data.data import InputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_builder import PipelineBuilder
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from golem.core.tuning.simultaneous import SimultaneousTuner
from matplotlib import pyplot as plt

from fedot_ind.api.utils.path_lib import default_path_to_save_results
from fedot_ind.api.utils.saver_collections import ResultSaver
from fedot_ind.core.metrics.evaluation import PerformanceAnalyzer
from fedot_ind.core.operation.transformation.splitter import TSTransformer
from fedot_ind.core.repository.initializer_industrial_models import IndustrialModels

np.random.seed(0)


class TimeSeriesAnomalyDetectionPreset:
    """Class responsible for interaction with Fedot classifier. It allows to use FEDOT optimization
    for hyperparameters tuning and pipeline building. Nodes of the pipeline are basis functions
    from the list of branch_nodes and statistical extractor.

    Attributes:
        branch_nodes: list of nodes to be used in the pipeline
        tuning_iters: number of iterations for tuning hyperparameters of preprocessing pipeline
        model_params: parameters of the FEDOT classification model
        dataset_name: name of the dataset to be used
        output_dir: path to the directory where results will be saved
        saver: object of ``ResultSaver`` class

    Notes: ``branch_nodes`` can be one or combination of the following: ``'data_driven_basis'``, ``'fourier_basis'``,
           ``'wavelet_basis'``.

    """

    def __init__(self, params: Optional[OperationParameters] = None):
        self.branch_nodes: list = params.get('branch_nodes', None)

        self.model_params = params.get('model_params')
        self.dataset_name = params.get('dataset')
        self.tuning_iters = params.get('tuning_iterations', 30)
        self.tuning_timeout = params.get('tuning_timeout', 15.0)
        self.output_folder = params.get('output_folder', default_path_to_save_results())

        self.saver = ResultSaver(dataset_name=self.dataset_name,
                                 generator_name='fedot_preset',
                                 output_dir=self.output_folder)
        self.logger = logging.getLogger('TimeSeriesAnomalyDetection_Preset')

        self.prediction_label = None
        self.auto_model = None
        self.y_train = None
        self.train_features = None
        self.test_features = None
        self.input_test_data = None
        self.generator = self._build_pipeline()
        self.predictor = None

        self.logger.info(f'TimeSeriesClassifierPreset initialised with [{self.branch_nodes}] nodes and '
                         f'[{self.tuning_iters}] tuning iterations and [{self.tuning_timeout}] timeout')

    def __check_multivariate_data(self, series: [np.array, List]) -> bool:
        """Method for checking if the data is multivariate.

        Args:
            series: time series

        Returns:
            True if data is multivariate, False otherwise

        """
        return isinstance(series, list) or (len(series.shape) > 1 and series.shape[1] > 1)

    def _init_input_data(self, series: np.array, anomaly_dict: Optional[Dict] = None, is_fit_stage=True) -> InputData:
        """Method for initialization of InputData object from pandas DataFrame and numpy array with target values.

        Args:
            series: numpy array with time series
            anomaly_dict: Dict with anomaly intervals
            is_fit_stage: is it fit stage or not

        Returns:
            InputData object convenient for FEDOT framework

        """
        if is_fit_stage:
            self.splitter = TSTransformer(
                strategy='frequent')
            features, target = self.splitter.transform_for_fit(series=series,
                                                               anomaly_dict=anomaly_dict,
                                                               plot=False,
                                                               binarize=False)
        else:
            features = self.splitter.transform(series=series)
            target = None

        is_multivariate_data = self.__check_multivariate_data(features)
        if is_multivariate_data:
            input_data = InputData(idx=np.arange(len(features)),
                                   features=np.array(features.tolist()),
                                   target=target.reshape(-1, 1) if target is not None else None,
                                   task=Task(TaskTypesEnum.classification),
                                   data_type=DataTypesEnum.image)
        else:
            input_data = InputData(idx=np.arange(len(features)),
                                   features=features.values,
                                   target=np.ravel(target).reshape(-1, 1) if target is not None else None,
                                   task=Task(TaskTypesEnum.classification),
                                   data_type=DataTypesEnum.table)

        return input_data

    def _build_pipeline(self):
        """
        Method for building pipeline with nodes from ``branch_nodes`` list and statistical extractor.

        """
        if self.branch_nodes is None:
            self.branch_nodes = ['eigen_basis', 'fourier_basis', 'wavelet_basis']

        self.extractors = ['quantile_extractor'] * len(self.branch_nodes)

        pipeline_builder = PipelineBuilder()

        for index, (basis, extractor) in enumerate(zip(self.branch_nodes, self.extractors)):
            pipeline_builder.add_node(basis, branch_idx=index)
            pipeline_builder.add_node(extractor, branch_idx=index)
        pipeline_builder.join_branches('mlp', params={'hidden_layer_sizes': (256, 128, 64, 32),
                                                      'max_iter': 300,
                                                      'activation': 'relu',
                                                      'solver': 'adam', })

        return pipeline_builder.build()

    def _tune_pipeline(self, pipeline: Pipeline, train_data: InputData):
        """
        Method for tuning pipeline with simultaneous tuner.

        Args:
            pipeline: pipeline to be tuned
            train_data: InputData object with train data

        Returns:
            tuned pipeline
        """
        if train_data.num_classes > 2:
            metric = ClassificationMetricsEnum.f1
        else:
            metric = ClassificationMetricsEnum.ROCAUC

        pipeline_tuner = TunerBuilder(train_data.task) \
            .with_tuner(SimultaneousTuner) \
            .with_metric(metric) \
            .with_timeout(self.tuning_timeout) \
            .with_iterations(self.tuning_iters) \
            .build(train_data)

        pipeline = pipeline_tuner.tune(pipeline)
        return pipeline

    def fit(self, features,
            anomaly_dict: List = None) -> object:
        """
        Method for fitting pipeline on train data. It also tunes pipeline and updates it with categorical features.

        Args:
            features: pandas DataFrame with features
            anomaly_dict: numpy array with anomaly intervals

        Returns:
            fitted FEDOT model as object of ``Pipeline`` class

        """

        with IndustrialModels():
            self.train_data = self._init_input_data(features, anomaly_dict)
            self.generator = self._tune_pipeline(self.generator,
                                                 self.train_data)
            self.generator.fit(self.train_data)
            train_data_preprocessed = self.generator.root_node.predict(self.train_data)
            train_data_preprocessed.predict = np.squeeze(train_data_preprocessed.predict)

            train_data_preprocessed = InputData(idx=train_data_preprocessed.idx,
                                                features=train_data_preprocessed.predict,
                                                target=train_data_preprocessed.target,
                                                data_type=train_data_preprocessed.data_type,
                                                task=train_data_preprocessed.task)

        metric = 'roc_auc' if train_data_preprocessed.num_classes == 2 else 'f1'
        self.model_params.update({'metric': metric})
        self.auto_model = Fedot(available_operations=['scaling',
                                                      'normalization',
                                                      'fast_ica',
                                                      'xgboost',
                                                      'rfr',
                                                      'rf',
                                                      'logit',
                                                      'mlp',
                                                      'knn',
                                                      'lgbm',
                                                      'pca'],
                                **self.model_params)

        self.auto_model.fit(train_data_preprocessed)
        self.predictor = self.auto_model.current_pipeline
        return self.auto_model

    def predict(self, features: pd.DataFrame) -> np.array:
        """
        Method for prediction on test data.

        Args:
            features: pandas DataFrame with features

        """

        test_data = self._init_input_data(features, is_fit_stage=False)
        test_data_preprocessed = self.generator.root_node.predict(test_data)

        if test_data.features.shape[0] == 1:
            test_data_preprocessed.predict = np.squeeze(test_data_preprocessed.predict).reshape(1, -1)
        else:
            test_data_preprocessed.predict = np.squeeze(test_data_preprocessed.predict)
        test_data_preprocessed = InputData(idx=test_data_preprocessed.idx,
                                           features=test_data_preprocessed.predict,
                                           target=test_data_preprocessed.target,
                                           data_type=test_data_preprocessed.data_type,
                                           task=test_data_preprocessed.task)

        self.prediction_label = self.auto_model.predict(test_data_preprocessed)
        self.prediction_label = self.convert_to_point_prediction(self.prediction_label, features)
        return self.prediction_label

    def predict_proba(self, features) -> np.array:
        test_data = self._init_input_data(features, is_fit_stage=False)
        test_data_preprocessed = self.generator.predict(test_data)
        test_data_preprocessed.predict = np.squeeze(test_data_preprocessed.predict)
        test_data_preprocessed = InputData(idx=test_data_preprocessed.idx,
                                           features=test_data_preprocessed.predict,
                                           target=test_data_preprocessed.target,
                                           data_type=test_data_preprocessed.data_type,
                                           task=test_data_preprocessed.task)
        self.prediction_proba = self.auto_model.predict_proba(test_data_preprocessed)
        self.prediction_proba = self.convert_to_point_prediction(self.prediction_proba, features, output_type='probs')
        return self.prediction_proba

    def get_metrics(self, target: Union[np.ndarray, pd.Series], metric_names: Union[str, List[str]]) -> dict:
        """
        Method for calculating metrics on test data.

        Args:
            target: numpy array with target values
            metric_names: list of desirable metrics names

        Returns:
            dictionary with metrics values that looks like ``{metric_name: metric_value}``

        """
        analyzer = PerformanceAnalyzer()
        return analyzer.calculate_metrics(target=np.ravel(target),
                                          predicted_labels=self.prediction_label,
                                          predicted_probs=self.prediction_proba,
                                          target_metrics=metric_names)

    def save_prediction(self, predicted_data: np.ndarray, kind: str):
        self.saver.save(predicted_data, kind)

    def save_metrics(self, metrics: dict):
        self.saver.save(metrics, 'metrics')

    def convert_to_point_prediction(self, prediction: np.array, features: np.array,
                                    output_type: str = 'labels') -> np.array:
        """
        Converts window-like prediction to point-like prediction

        :param prediction: window-like prediction
        :param features: features array. Must have equal length to target.
        :param output_type: output_type of prediction. Possible values are "labels" or "probs"
        """
        predicted = []
        for point in range(len(features)):
            if output_type == 'labels':
                pred_for_point = prediction[point // self.splitter.freq_length].item()
            else:
                pred_for_point = prediction[point // self.splitter.freq_length]
            predicted.append(pred_for_point)
        return predicted

    def save(self, path: Path):
        """
        Saves generator and predictor pipelines in 'path' folder. Generator saves into path 'path'/generator,
        predictor into path 'path'/predictor

        :param path: path to save pipelines
        """
        self.generator.save(Path(path, 'generator'))
        self.predictor.save(Path(path, 'predictor'))

    def load(self, path: Path):
        """
        Loads generator and predictor pipelines from 'path' folder. Generator loads from path 'path'/generator,
        predictor loads from 'path'/predictor

        :param path: path to load pipelines
        """
        with IndustrialModels():
            self.generator = Pipeline().load(Path(path, 'generator', '0_pipeline_saved', '0_pipeline_saved.json'))
        self.predictor = Pipeline().load(Path(path, 'predictor', '0_pipeline_saved', '0_pipeline_saved.json'))

    def plot_prediction(self, series, true_anomalies):
        fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(15, 7))
        for i in ax:
            i.plot(series, color='black')
            i.set_xlabel('Time')
            i.set_ylabel('Value')
        ax[0].set_title('Real anomalies')
        ax[1].set_title('Predicted anomalies')

        unique_anomalies = np.unique(true_anomalies)
        unique_anomalies = np.setdiff1d(unique_anomalies, np.array(['no_anomaly']))

        def generate_colors(num_colors):
            colormap = plt.cm.get_cmap('tab10')
            colors = []
            for i in range(num_colors):
                colors.append(colormap(i))
            return colors

        cmap = generate_colors(len(unique_anomalies))
        color_dict = {cls: color for cls, color in zip(unique_anomalies, cmap)}

        legend_patches = [mpatches.Patch(color=color_dict[cls],
                                         label=cls) for cls in unique_anomalies]

        for anomaly_class in unique_anomalies:
            ax[0].fill_between(np.arange(len(series)), np.min(series),
                               np.max(series),
                               where=true_anomalies == anomaly_class,
                               facecolor=color_dict[anomaly_class], alpha=.4)
            ax[1].fill_between(np.arange(len(series)), np.min(series),
                               np.max(series),
                               where=np.array(self.prediction_label) == anomaly_class,
                               facecolor=color_dict[anomaly_class], alpha=.4)

        plt.legend(handles=set(legend_patches))
        plt.show()



"""This module contains classes for object detection task based on torch dataset."""

import os
import json
import yaml
from typing import Tuple, Callable, Dict

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset
from tqdm import tqdm

IMG_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif', '.tiff', '.webp')


class COCODataset(Dataset):
    """Class-loader for COCO json.

    Args:
        images_path: Image folder path.
        json_path: Json file path.
        transform: A function/transform that takes in an PIL image and returns a
            transformed version.
        fix_zero_class: If ``True`` add 1 for each class label
            (0 represents always the background class).
        replace_to_binary: If ``True`` set label 1 for any class.
    """

    def __init__(
        self,
        images_path: str,
        json_path: str,
        transform: Callable,
        fix_zero_class: bool = False,
        replace_to_binary: bool = False,
    ) -> None:
        self.transform = transform
        self.classes = {}
        self.samples = []

        with open(json_path) as f:
            data = json.load(f)

        for category in data['categories']:
            id = category['id'] + 1 if fix_zero_class else category['id']
            self.classes[id] = category['name']

        samples = {}
        for image in data['images']:
            samples[image['id']] = {
                'image': os.path.join(images_path, image['file_name']),
                'area': [],
                'iscrowd': [],
                'labels': [],
                'boxes': [],
            }

        for annotation in tqdm(data['annotations']):
            if annotation['area'] > 0:
                bbox = np.array(annotation['bbox'])
                bbox[2:] += bbox[:2]  # x, y, w, h -> x1, y1, x2, y2
                labels = annotation['category_id']
                labels = labels + 1 if fix_zero_class else labels
                labels = 1 if replace_to_binary else labels
                samples[annotation['image_id']]['labels'].append(labels)
                samples[annotation['image_id']]['boxes'].append(bbox)
                samples[annotation['image_id']]['area'].append(annotation['area'])
                samples[annotation['image_id']]['iscrowd'].append(annotation['iscrowd'])

        for sample in samples.values():
            if len(sample['labels']) > 0:
                self.samples.append(sample)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Returns a sample from a dataset.

        Args:
            idx: Index of sample.

        Returns:
            A tuple ``(image, targets)``, where image is image tensor,
                and targets is dict with keys: ``'boxes'``, ``'labels'``,
                ``'image_id'``, ``'area'``, ``'iscrowd'``.
        """
        sample = self.samples[idx]
        image = Image.open(sample['image']).convert('RGB')
        image = self.transform(image)
        target = {
            'boxes': torch.tensor(np.stack(sample['boxes']), dtype=torch.float32),
            'labels': torch.tensor(sample['labels'], dtype=torch.int64),
            'image_id': torch.tensor([idx]),
            'area': torch.tensor(sample['area'], dtype=torch.float32),
            'iscrowd': torch.tensor(sample['iscrowd'], dtype=torch.int64),
        }
        return image, target

    def __len__(self) -> int:
        """Return length of dataset"""
        return len(self.samples)


def img2label_paths(img_path):
    """Define label path as a function of image path."""
    sa, sb = f'{os.sep}images{os.sep}', f'{os.sep}labels{os.sep}'  # /images/, /labels/ substrings
    return sb.join(img_path.rsplit(sa, 1)).rsplit('.', 1)[0] + '.txt'


class YOLODataset(Dataset):
    """Class-loader for YOLO format (https://docs.ultralytics.com/datasets/detect/).

    Args:
        path: YAML file path.
        transform: A function/transform that takes in an PIL image and returns a
            transformed version.
        train: If True, creates dataset from training set, otherwise creates from test set.
        replace_to_binary: If ``True`` set label 1 for any class.
    """

    def __init__(
        self,
        path: str,
        transform: Callable,
        train: bool = True,
        replace_to_binary: bool = False,
    ) -> None:

        self.transform = transform
        with open(path, 'r') as f:
            data = yaml.safe_load(f)
        self.root = os.path.abspath(os.path.join(os.path.dirname(path), (data['train'] if train else data['val'])))
        self.classes = ['background']
        self.classes.extend(['object'] if replace_to_binary else data['names'])
        self.binary = replace_to_binary
        self.samples = []

        for file in os.listdir(self.root):
            if file.lower().endswith(IMG_EXTENSIONS):
                self.samples.append(
                    {
                        'image': os.path.join(self.root, file),
                        'label': img2label_paths(os.path.join(self.root, file))
                    }
                )

    def __getitem__(self, idx) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Returns a sample from a dataset.

        Args:
            idx: Index of sample.

        Returns:
            A tuple ``(image, targets)``, where image is image tensor,
                and targets is dict with keys: ``'boxes'``, ``'labels'``,
                ``'image_id'``, ``'area'``, ``'iscrowd'``.
        """
        sample = self.samples[idx]
        image = Image.open(sample['image']).convert('RGB')
        image = self.transform(image)
        annotation = np.loadtxt(sample['label'], ndmin=2)
        labels = annotation[:, 0] + 1
        labels = np.ones_like(labels) if self.binary else labels
        boxes = annotation[:, 1:]
        c, h, w = image.shape
        boxes *= [w, h, w, h]
        area = boxes[:, 2] * boxes[:, 3]
        boxes[:, :2] -= boxes[:, 2:] / 2  # x centre, y centre, w, h -> x1, y1, w, h
        boxes[:, 2:] += boxes[:, :2]  # x1, y1, w, h -> x1, y1, x2, y2

        target = {
            'boxes': torch.tensor(boxes, dtype=torch.float32),
            'labels': torch.tensor(labels, dtype=torch.int64),
            'image_id': torch.tensor([idx]),
            'area': torch.tensor(area, dtype=torch.float32),
            'iscrowd': torch.zeros(annotation.shape[0], dtype=torch.int64),
        }
        return image, target

    def __len__(self) -> int:
        """Return length of dataset"""
        return len(self.samples)



from typing import Union

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from fedot_ind.tools.synthetic.anomalies import AddNoise, DecreaseDispersion, Dip, IncreaseDispersion, Peak, \
    ShiftTrendDOWN, \
    ShiftTrendUP
from fedot_ind.tools.synthetic.ts_generator import TimeSeriesGenerator


class AnomalyGenerator:
    """
    AnomalyGenerator class is used to generate anomalies in time series data. It takes time series data as input and
    returns time series data with anomalies. Anomalies are generated based on anomaly_config parameter. It is a dict
    with anomaly class names as keys and anomaly parameters as values. Anomaly class names must be the same as anomaly
    class names in anomalies.py.

    Attributes:
        types: dict. Dict with anomaly class names as keys and anomaly parameters as values.
        anomaly_config: dict, default=BASE_CONFIG. Dict with anomaly class names as keys and anomaly parameters as
            values.
        taken_slots: np.ndarray, default=None. Array of 0 and 1. 1 means that this time slot is already taken by
            another anomaly.
        overlap: float, default=0.1. Argument of `generate` method. Defines the maximum overlap between anomalies.

    Example:
        First, we need to create an instance of AnomalyGenerator class with config as its argument where every anomaly
        type hyperparameters are defined::
            config = {'add_noise': {'level': 80,
                                    'number': 6,
                                    'noise_type': 'gaussian',
                                    'min_anomaly_length': 10,
                                    'max_anomaly_length': 20}}
            generator = AnomalyGenerator(config=config)

        Then we can generate anomalies in time series data using method `generate` which arguments are
        `time_series_data` (`np.array` of config for synthetic ts_data), `plot` and acceptable `overlap`::
            initial_ts, modified_ts, intervals = generator.generate(time_series_data=data,
                                                                    plot=True,
                                                                    overlap=0.1)

        This method returns initial time series data, modified time series data and dict with anomaly intervals.

    """

    def __init__(self, **params):
        self.types = {
            'decrease_dispersion': DecreaseDispersion,
            'increase_dispersion': IncreaseDispersion,
            'shift_trend_up': ShiftTrendUP,
            'shift_trend_down': ShiftTrendDOWN,
            'add_noise': AddNoise,
            'dip': Dip,
            'peak': Peak}

        self.anomaly_config = params.get('config', ValueError('config must be defined'))
        self.taken_slots = None
        self.overlap = None

    def select_interval(self, max_length: int, min_length: int) -> tuple:
        ts_length = self.taken_slots.size
        start_idx = np.random.randint(max_length, ts_length - max_length)
        end_idx = start_idx + np.random.randint(min_length, max_length + 1)

        if self.taken_slots[start_idx:end_idx].mean() > self.overlap:
            return self.select_interval(max_length, min_length)
        else:
            self.taken_slots[start_idx:end_idx] = 1
            return start_idx, end_idx

    def generate(self, time_series_data: Union[np.ndarray, dict], plot: bool = False, overlap: float = 0.1):
        """
        Generate anomalies in time series data.

        Args:
            time_series_data: either np.ndarray or dict with config for synthetic ts_data.
            plot: if True, plot initial and modified time series data with rectangle spans of anomalies.
            overlap: float, ``default=0.1``. Defines the maximum overlap between anomalies.

        Returns:
            returns initial time series data, modified time series data and dict with anomaly intervals.

        """
        if isinstance(time_series_data, dict):
            ts_generator = TimeSeriesGenerator(time_series_data)
            t_series = ts_generator.get_ts()

        elif isinstance(time_series_data, np.ndarray):
            t_series = time_series_data
        else:
            raise ValueError('time_series_data must be np.ndarray or dict')

        initial_ts = t_series.copy()
        anomaly_intervals_dict = {}

        self.taken_slots = pd.Series([0 for _ in t_series])
        self.overlap = overlap

        for anomaly_cls in self.anomaly_config.keys():
            n = self.anomaly_config[anomaly_cls]['number']
            anomaly_obj = self.types[anomaly_cls]
            params = self.anomaly_config[anomaly_cls]
            max_length = params.get('max_anomaly_length',
                                    ValueError(f'max_anomaly_length must be defined for {anomaly_cls} type'))
            min_length = params.get('min_anomaly_length',
                                    ValueError(f'min_anomaly_length must be defined for {anomaly_cls} type'))

            for i in range(n):
                start_idx, end_idx = self.select_interval(max_length, min_length)
                t_series = anomaly_obj(params).get(ts=t_series, interval=(start_idx, end_idx))

                if anomaly_cls in anomaly_intervals_dict:
                    anomaly_intervals_dict[anomaly_cls].append([start_idx, end_idx])
                else:
                    anomaly_intervals_dict[anomaly_cls] = [[start_idx, end_idx]]

        if plot:
            self.plot_anomalies(initial_ts=initial_ts, modified_ts=t_series,
                                anomaly_intervals_dict=anomaly_intervals_dict)

        return initial_ts, t_series, anomaly_intervals_dict

    def plot_anomalies(self, initial_ts, modified_ts, anomaly_intervals_dict):
        fig, ax = plt.subplots(figsize=(12, 7))
        ax.plot(modified_ts, label='Modified Time Series')
        ax.plot(initial_ts, label='Initial Time Series')
        ax.set_title('Time Series with Anomalies')
        ax.set_xlabel('Time')
        ax.set_ylabel('Value')

        cmap = self.generate_colors(len(anomaly_intervals_dict.keys()))
        color_dict = {cls: color for cls, color in zip(anomaly_intervals_dict.keys(), cmap)}

        legend_patches = [patches.Patch(color=color_dict[cls],
                                        label=cls) for cls in anomaly_intervals_dict.keys()]

        for anomaly_class, intervals in anomaly_intervals_dict.items():
            for interval in intervals.transform_for_fit(', '):
                start_idx, end_idx = map(int, interval.transform_for_fit(':'))
                ax.axvspan(start_idx, end_idx, alpha=0.3, color=color_dict[anomaly_class])

        # Put a legend to the right of the current axis
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), handles=set(legend_patches))
        plt.show()

    def generate_colors(self, num_colors: int) -> list:
        colormap = plt.cm.get_cmap('tab10')
        colors = [colormap(i) for i in range(num_colors)]
        return colors


if __name__ == '__main__':

    # config = {'decrease_dispersion': {'level': 70,
    #                                   'number': 2,
    #                                   'min_anomaly_length': 10,
    #                                   'max_anomaly_length': 15},
    #           'dip': {'level': 20,
    #                   'number': 2,
    #                   'min_anomaly_length': 10,
    #                   'max_anomaly_length': 20},
    #
    #           'peak': {'level': 2,
    #                    'number': 2,
    #                    'min_anomaly_length': 5,
    #                    'max_anomaly_length': 10},
    #           'increase_dispersion': {'level': 70,
    #                                   'number': 2,
    #                                   'min_anomaly_length': 30,
    #                                   'max_anomaly_length': 40},
    #           'shift_trend_up': {'level': 10,
    #                              'number': 2,
    #                              'min_anomaly_length': 10,
    #                              'max_anomaly_length': 20},
    #           'shift_trend_down': {'level': 10,
    #                                'number': 2,
    #                                'min_anomaly_length': 10,
    #                                'max_anomaly_length': 20},
    #           'add_noise': {'level': 80,
    #                         'number': 2,
    #                         'noise_type': 'uniform',
    #                         'min_anomaly_length': 50,
    #                         'max_anomaly_length': 60}
    #           }
    #
    # generator = AnomalyGenerator(config=config)
    #
    # ts_conf = {'ts_type': 'sin',
    #            'ts_length': 2000}
    #
    # init_ts, mot_ts, inters = generator.generate(time_series_data=ts_conf,
    #                                              plot=True,
    #                                              overlap=0.1)

    synth_ts = {'ts_type': 'sin',
                'length': 1000,
                'amplitude': 10,
                'period': 500}

    anomaly_config = {'dip': {'level': 20,
                              'number': 5,
                              'min_anomaly_length': 10,
                              'max_anomaly_length': 20},
                      'peak': {'level': 2,
                               'number': 5,
                               'min_anomaly_length': 5,
                               'max_anomaly_length': 10},
                      'decrease_dispersion': {'level': 70,
                                                'number': 2,
                                                'min_anomaly_length': 10,
                                                'max_anomaly_length': 15},
                      # 'increase_dispersion': {'level': 50,
                      #                         'number': 2,
                      #                         'min_anomaly_length': 10,
                      #                         'max_anomaly_length': 15},
                      # 'shift_trend_up': {'level': 10,
                      #                    'number': 2,
                      #                    'min_anomaly_length': 10,
                      #                    'max_anomaly_length': 20},
                      # 'add_noise': {'level': 80,
                      #               'number': 2,
                      #               'noise_type': 'uniform',
                      #               'min_anomaly_length': 10,
                      #               'max_anomaly_length': 20}
                      }

    generator = AnomalyGenerator(config=anomaly_config)

    init_synth_ts, mod_synth_ts, synth_inters = generator.generate(time_series_data=synth_ts,
                                                                   plot=True,
                                                                   overlap=0.1)
    _ = 1



import os
from typing import Dict, Union, Optional, List, Tuple, Callable
from functools import wraps
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import numpy as np
import pandas as pd


def savefig(func) -> Callable:
    """Adds the ability to save a figure."""

    @wraps(func)
    def wrapper(saving_path: Optional[Union[str, Path]] = None, **kwargs):
        fig = func(**kwargs)
        if saving_path is not None:
            fig.savefig(saving_path)
        return fig

    return wrapper


def limits(func) -> Callable:
    """Adds the ability to set plot parameters."""

    @wraps(func)
    def wrapper(
            title: str = '',
            xlabel: str = '',
            ylabel: str = '',
            xlim: Optional[Tuple[float, float]] = None,
            ylim: Optional[Tuple[float, float]] = None,
            **kwargs
    ):
        fig = func(**kwargs)
        plt.grid()
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)
        if xlim is not None:
            plt.xlim(xlim)
        if ylim is not None:
            plt.ylim(ylim)
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        return fig

    return wrapper


def random_color() -> Tuple[float, float, float]:
    """Generates a random color.

    Returns:
        Tuple: (r, g, b).
    """
    r = np.random.random_sample()
    g = np.random.random_sample()
    b = np.random.random_sample()
    return r, g, b


def pair_color(r: float, g: float, b: float, delta=0.2) -> Tuple[float, float, float]:
    """Generates close to the specified color.

    Args:
        r: The red component of the color from the range [0, 1].
        g: The green component of the color from the range [0, 1].
        b: The blue component of the color from the range [0, 1].
        delta: An offset relative to the specified color.

    Returns:
        Tuple: (r, g, b).
    """
    r = r + delta if r + delta <= 1 else 1
    g = g + delta if g + delta <= 1 else 1
    b = b + delta if b + delta <= 1 else 1
    return r, g, b


def get_best_metric(
        exp_path: Union[str, Path],
        metric: str,
        phase: str = 'train'
) -> float:
    """Returns the best metric score from training history.

    Args:
        exp_path: Path to the experiment metrics folder.
        metric: Target metric.
        phase: Experiment phase, e.g. `"train"`.

    Returns:
        The best metric score.
    """
    df = pd.read_csv(os.path.join(exp_path, f'{phase}/val.csv'), index_col=0)
    return df[metric].max()


def create_mean_exp(path: str) -> None:
    """
    Averages all experiments in a folder.

    Args:
        path: Path to the folder with the results of experiments.
    """
    exps = os.listdir(path)
    phases = os.listdir(os.path.join(path, exps[0]))
    os.mkdir(os.path.join(path, 'mean'))
    for phase in phases:
        if phase.endswith('.csv'):
            df = create_mean_df([os.path.join(path, exp, phase) for exp in exps])
            df.to_csv(os.path.join(path, 'mean', phase))
        elif os.path.isdir(os.path.join(path, exps[0], phase)):
            os.mkdir(os.path.join(path, 'mean', phase))
            train_df = create_mean_df([os.path.join(path, exp, phase, 'train.csv') for exp in exps])
            train_df.to_csv(os.path.join(path, 'mean', phase, 'train.csv'))
            val_df = create_mean_df([os.path.join(path, exp, phase, 'val.csv') for exp in exps])
            val_df.to_csv(os.path.join(path, 'mean', phase, 'val.csv'))


def create_mean_df(paths: List[Union[str, Path]]) -> pd.DataFrame:
    """
    Combine dataframes of experiments by calculating their mean and std.

    Args:
        paths: List of paths to dataframes.

    Returns:
        The dataframe of mean and std values.
    """
    df = pd.read_csv(paths[0], index_col=0)
    data = [df.to_numpy()]
    for path in paths[1:]:
        tmp = pd.read_csv(path, index_col=0)
        assert np.array_equal(df.index, tmp.index), f"{df.index} are not equal to {tmp.index} in {path}"
        assert np.array_equal(df.columns, tmp.columns), f"{df.columns} are not equal to {tmp.columns} in {path}"
        data.append(tmp.to_numpy())
    data = np.array(data)
    mean = data.mean(axis=0)
    std = data.std(axis=0)
    df = pd.DataFrame(
        data=np.concatenate((mean, std), axis=1),
        index=df.index,
        columns=df.columns.tolist() + [f'{c} std' for c in df.columns]
    )
    return df


@savefig
@limits
def show_train_scores(
        exps: Dict[str, Union[str, Path]],
        metric: str,
        show_std: bool = False,
        figsize: Tuple[int, int] = (16, 8)
) -> plt.Figure:
    """Draws mean training graphs.

    Args:
        exps: Dictionary of experiments: `{exp_name: [exp_paths]}`.
        metric: Target metric for plotting.
        show_std: If `True` displays standard deviation on graphs.
        figsize: Width, height in inches.

    Return:
        Figure with graphs.
    """
    fig = plt.figure(figsize=figsize)
    for name, exp in exps.items():
        exp_scores = pd.read_csv(os.path.join(exp, 'train/val.csv'), index_col=0)
        r, g, b = random_color()
        plt.plot(exp_scores.index, exp_scores[metric], label=name, color=(r, g, b))
        if show_std:
            plt.fill_between(
                exp_scores.index,
                exp_scores[metric] + exp_scores[f'{metric} std'],
                exp_scores[metric] - exp_scores[f'{metric} std'],
                color=(r, g, b, 0.3)
            )
    return fig


def compare_svd_results(
        baseline: Union[str, Path],
        svd_exps: Dict[str, Union[str, Path]],
        metric: str,
) -> Dict[str, pd.DataFrame]:
    """Creates dataframe.

    Args:
        baseline: The path to the base experiment for comparison.
        svd_exps: Dictionary of experiments: `{exp_name: exp_path}`.
        metric: Target metric.

    Returns:
        Dictionary `{exp_name: exp_dataframe}`.
    """
    factor = 100 / get_best_metric(exp_path=baseline, metric=metric)
    result = {}

    for exp, path in svd_exps.items():
        metrics_df = pd.read_csv(os.path.join(path, 'pruning.csv'), index_col=0)
        size_df = pd.read_csv(os.path.join(path, 'size.csv'), index_col=0)
        df = pd.DataFrame()
        df['size'] = metrics_df['size'] / size_df.loc['default', 'size'] * 100
        df['pruned'] = metrics_df[metric] * factor
        if f'{metric} std' in metrics_df.columns:
            df['pruned std'] = metrics_df[f'{metric} std'] * factor

        for e in df.index:
            e_path = os.path.join(path, e, 'val.csv')
            if os.path.exists(e_path):
                e_df = pd.read_csv(e_path, index_col=0)
                idx = e_df[metric].idxmax()
                df.loc[e, 'fine-tuned'] = e_df.loc[idx, metric] * factor
                if f'{metric} std' in e_df.columns:
                    df['fine-tuned std'] = e_df.loc[idx, f'{metric} std'] * factor
        result[exp] = df
    return result


@savefig
@limits
def show_svd_results(
        svd_exps: Dict[str, pd.DataFrame],
        figsize: Tuple[int, int] = (16, 8),
        fig: Optional[plt.Figure] = None,
) -> plt.Figure:
    """Draws graphs of the mean target metric depending on the mean size of the compressed SVD models.

    Args:
        svd_exps: Dictionary of experiments: `{exp_name: exp_dataframe}`
        fig: Figure for drawing graphs.
        figsize: Width, height in inches.

    Returns:
        Figure with graphs.
    """
    if fig is None:
        fig, ax = plt.subplots(figsize=figsize)

    for exp, df in svd_exps.items():
        r, g, b = random_color()

        if 'pruned' in df.columns:
            plt.plot(df['size'], df['pruned'], label=exp, color=(r, g, b))
            if 'pruned std' in df.columns:
                plt.fill_between(
                    df['size'],
                    df['pruned'] + df['pruned std'],
                    df['pruned'] - df['pruned std'],
                    color=(r, g, b, 0.3)
                )

        if 'fine-tuned' in df.columns:
            plt.plot(df['size'], df['fine-tuned'], label=f'{exp} fine-tuned', color=pair_color(r, g, b))
            if 'fine-tuned std' in df.columns:
                plt.fill_between(
                    df['size'],
                    df['fine-tuned'] + df['fine-tuned std'],
                    df['fine-tuned'] - df['fine-tuned std'],
                    color=(*pair_color(r, g, b), 0.3),
                )
    return fig


def compare_sfp_results(
        baseline: Union[str, Path],
        sfp_exps: Dict[str, Union[str, Path]],
        metric: str,
) -> pd.DataFrame:
    """Creates dataframe.

    Args:
        baseline: The path to the base experiment for comparison.
        sfp_exps: Dictionary of experiments: `{exp_name: exp_path}`.
        metric: Target metric.

    Returns:
        Dataframe.
    """
    factor = 100 / get_best_metric(exp_path=baseline, metric=metric)
    df = pd.DataFrame()
    for exp, path in sfp_exps.items():
        metric_df = pd.read_csv(os.path.join(path, 'train/val.csv'), index_col=0)
        size_df = pd.read_csv(os.path.join(path, 'size.csv'), index_col=0)
        size_factor = 100 / size_df.loc['default', 'size']
        df.loc[exp, 'size'] = size_df.loc['pruned', 'size'] * size_factor

        if 'size std' in size_df.columns:
            df.loc[exp, 'size std'] = size_df.loc['pruned', 'size std'] * size_factor

        idx = metric_df[metric].idxmax()
        df.loc[exp, 'pruned'] = metric_df.loc[idx, metric] * factor
        if f'{metric} std' in metric_df.columns:
            df.loc[exp, 'pruned std'] = metric_df.loc[idx, f'{metric} std'] * factor
        ft_path = os.path.join(path, 'pruned/val.csv')
        if os.path.exists(ft_path):
            ft_df = pd.read_csv(ft_path, index_col=0)
            idx = ft_df[metric].idxmax()
            df.loc[exp, 'fine-tuned'] = ft_df.loc[idx, metric] * factor
            if f'{metric} std' in ft_df.columns:
                df.loc[exp, 'fine-tuned std'] = ft_df.loc[idx, f'{metric} std'] * factor
    return df


@savefig
@limits
def show_sfp_results(
        sfp_exps: pd.DataFrame,
        figsize: Tuple[int, int] = (16, 8),
        fig: Optional[plt.Figure] = None,
):
    """Draws graphs of the mean target metric depending on the mean size of the compressed SFP models.

    Args:
        sfp_exps: Dataframe of experiments.
        fig: Figure for drawing graphs.
        figsize: Width, height in inches.

    Returns:
        Figure with graphs.
    """
    if fig is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        ax = fig.axes[0]
    for exp in sfp_exps.index:
        r, g, b = random_color()
        if 'pruned' in sfp_exps.columns:
            plt.scatter(
                sfp_exps.loc[exp, 'size'],
                sfp_exps.loc[exp, 'pruned'],
                label=exp,
                color=(r, g, b),
                marker='o',
            )
            if 'pruned std' in sfp_exps.columns:
                ax.add_patch(Rectangle(
                    xy=(
                        sfp_exps.loc[exp, 'size'] - sfp_exps.loc[exp, 'size std'],
                        sfp_exps.loc[exp, 'pruned'] - sfp_exps.loc[exp, 'pruned std'],
                    ),
                    width=sfp_exps.loc[exp, 'size std'] * 2,
                    height=sfp_exps.loc[exp, 'pruned std'] * 2,
                    color=(r, g, b, 0.3),
                ))

        if 'fine-tuned' in sfp_exps.columns:
            plt.scatter(
                sfp_exps.loc[exp, 'size'],
                sfp_exps.loc[exp, 'fine-tuned'],
                label=f'{exp} fine-tuned',
                color=pair_color(r, g, b),
                marker='o',
            )
            if 'fine-tuned std' in sfp_exps.columns:
                ax.add_patch(Rectangle(
                    xy=(
                        sfp_exps.loc[exp, 'size'] - sfp_exps.loc[exp, 'size std'],
                        sfp_exps.loc[exp, 'fine-tuned'] - sfp_exps.loc[exp, 'fine-tuned std'],
                    ),
                    width=sfp_exps.loc[exp, 'size std'] * 2,
                    height=sfp_exps.loc[exp, 'fine-tuned std'] * 2,
                    color=(*pair_color(r, g, b), 0.3),
                ))
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    return fig


@savefig
@limits
def show_svd_sfp_results(
        sfp_exps: pd.DataFrame,
        svd_exps: Dict[str, pd.DataFrame],
        figsize: Tuple[int, int] = (16, 8),
):
    """Draws graphs of the mean target metric depending on the mean size of the compressed SVD and SFP models.

    Args:
        sfp_exps: Dataframe of experiments.
        svd_exps: Dictionary of experiments: `{exp_name: exp_dataframe}`.
        figsize: Width, height in inches.

    Returns:
        Figure with graphs.
    """
    fig, ax = plt.subplots(figsize=figsize)
    show_svd_results(svd_exps=svd_exps, fig=fig)
    show_sfp_results(sfp_exps=sfp_exps, fig=fig)
    return fig



"""This module contains the class and functions for integrating the computer vision module into the framework API."""
import os
from typing import Callable, Dict, Optional, Tuple
from functools import partial
import logging
from urllib.error import URLError

import torch
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torchvision.datasets import ImageFolder
from torchvision.models import resnet18
from torchvision.models.detection import ssdlite320_mobilenet_v3_large
from torchvision.models.segmentation import deeplabv3_resnet50
from torchvision.transforms import ToTensor

from fedot_ind.core.architecture.abstraction.сheckers import parameter_value_check
from fedot_ind.core.architecture.datasets.object_detection_datasets import YOLODataset
from fedot_ind.core.architecture.datasets.prediction_datasets import PredictionFolderDataset
from fedot_ind.core.architecture.datasets.splitters import train_test_split
from fedot_ind.core.architecture.experiment.nn_experimenter import ClassificationExperimenter, FitParameters, \
    NNExperimenter, ObjectDetectionExperimenter, SegmentationExperimenter
from fedot_ind.core.operation.optimization.structure_optimization import SFPOptimization, StructureOptimization, \
    SVDOptimization


def get_classification_dataloaders(
        dataset_path: str,
        dataloader_params: Dict,
        transform: Callable
) -> Tuple[DataLoader, DataLoader, Dict[int, str]]:
    """
    Returns the training and validation data loaders for image classification (`torchvision.datasets.ImageFolder`).

    Args:
        dataset_path: Image folder path.
        dataloader_params: Parameter dictionary passed to `torch.utils.data.DataLoader`
        transform: The image transformation function passed to (`torchvision.datasets.ImageFolder`).

    Returns:
        `(train_dataloader, validation_dataloader, idx_to_class)`
    """
    ds = ImageFolder(root=dataset_path, transform=transform)
    train_ds, val_ds = train_test_split(ds)
    train_dl = DataLoader(train_ds, shuffle=True, **dataloader_params)
    val_dl = DataLoader(val_ds, **dataloader_params)
    idx_to_class = {idx: cls for cls, idx in ds.class_to_idx.items()}
    return train_dl, val_dl, idx_to_class


def get_object_detection_dataloaders(
        dataset_path: str,
        dataloader_params: Dict,
        transform: Callable
) -> Tuple[DataLoader, DataLoader, Dict[int, str]]:
    """
    Returns the training and validation data loaders for object detection (YOLO style).

    Args:
        dataset_path: Path to folder with images and their annotations.
        dataloader_params: Parameter dictionary passed to `torch.utils.data.DataLoader`
        transform: The image transformation function passed to dataset initialization.

    Returns:
        `(train_dataloader, validation_dataloader, idx_to_class)`
    """
    train_ds = YOLODataset(path=dataset_path, transform=transform)
    val_ds = YOLODataset(path=dataset_path, transform=transform, train=False)
    train_dl = DataLoader(train_ds, shuffle=True, collate_fn=lambda x: tuple(zip(*x)), **dataloader_params)
    val_dl = DataLoader(val_ds, collate_fn=lambda x: tuple(zip(*x)), **dataloader_params)
    idx_to_class = {idx: cls for idx, cls in enumerate(train_ds.classes)}
    return train_dl, val_dl, idx_to_class


def get_segmentation_dataloaders(
        dataset_path: str,
        dataloader_params: Dict,
        transform: Callable
) -> Tuple[DataLoader, DataLoader]:
    return NotImplementedError


def classification_idx_to_class(
        preds: Dict[str, int],
        idx_to_class: Dict[int, str],
        proba: bool = False
) -> Dict[str, str]:
    return preds if proba else {img: idx_to_class[idx] for img, idx in preds.items()}


def detection_idx_to_class(
        preds: Dict[str, Dict],
        idx_to_class: Dict[int, str],
        proba: bool = False,
) -> Dict[str, Dict]:
    return {
        img: {
            k: (v if k != 'labels' else [idx_to_class[idx] for idx in v]) for k, v in pred.items()
        } for img, pred in preds.items()
    }


CV_TASKS = {
    'image_classification': {
        'experimenter': ClassificationExperimenter,
        'model': resnet18,
        'data': get_classification_dataloaders,
        'idx_to_class': classification_idx_to_class,
    },
    'object_detection': {
        'experimenter': ObjectDetectionExperimenter,
        'model': ssdlite320_mobilenet_v3_large,
        'data': get_object_detection_dataloaders,
        'idx_to_class': detection_idx_to_class,
    },
    'semantic_segmentation': {
        'experimenter': SegmentationExperimenter,
        'model': deeplabv3_resnet50,
        'data': get_segmentation_dataloaders,
        'idx_to_class': lambda x, y, z: x
    }
}

OPTIMIZATIONS = {
    'svd': SVDOptimization,
    'sfp': SFPOptimization,
}


class CVExperimenter:
    """
    This class is used to integrate the computer vision module into the framework API.

    Args:
        kwargs: Keyword arguments for the experimenter object.

    Attributes:
        task: Computer vision task name.
        path: Path to the folder where the results of the experiment will be saved.
        optim: Optimization object.
        exp: Experimenter object.
        val_dl: Validation data loader.

    """
    def __init__(self, kwargs):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.task: str = kwargs.pop('task')
        self.path: str = kwargs.pop('output_folder')
        self.optim: Optional[StructureOptimization] = None
        if 'model' in kwargs.keys():
            assert isinstance(kwargs['model'], torch.nn.Module), 'Model must be an instance of torch.nn.Module'
        else:
            assert 'num_classes' in kwargs.keys(), 'It is necessary to pass the number of classes or the model object'

            num_classes = kwargs.pop('num_classes')

            try:
                kwargs['model'] = CV_TASKS[self.task]['model'](num_classes=num_classes)
            except URLError:
                # Fix for possible SSL error
                import ssl
                ssl._create_default_https_context = ssl._create_unverified_context
                kwargs['model'] = CV_TASKS[self.task]['model'](num_classes=num_classes)

        if 'optimization' in kwargs.keys():
            optimization = kwargs.pop('optimization')
            parameter_value_check('optimization', optimization, set(OPTIMIZATIONS.keys()))
            opt_params = kwargs.pop('optimization_params', {})
            self.optim = OPTIMIZATIONS[optimization](**opt_params)

        if 'device' not in kwargs.keys():
            kwargs['device'] = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.exp: NNExperimenter = CV_TASKS[self.task]['experimenter'](**kwargs)
        self.val_dl: Optional[DataLoader] = None
        self.idx_to_class: Optional[Dict[int: str]] = None
        self.logger.info(f'{type(self.exp).__name__} initialised')

    def fit(self, dataset_path: str, **kwargs):
        """
        Starts training the model.

        Args:
            dataset_path: Path to dataset.

        Returns:
            Trained model (`torch.nn.Module`).
        """
        self.logger.info('Dataset preparing')
        train_dl, val_dl, idx_to_class = CV_TASKS[self.task]['data'](
            dataset_path=dataset_path,
            dataloader_params=kwargs.pop('dataloader_params', {'batch_size': 8, 'num_workers': 4}),
            transform=kwargs.pop('transform', ToTensor())
        )
        self.val_dl = val_dl
        self.idx_to_class = idx_to_class

        ds_name = kwargs.pop('dataset_name', dataset_path.split('/')[-1])
        ft_params = kwargs.pop('finetuning_params', {})
        num_epoch = kwargs.pop('num_epochs', 50)
        if 'lr_scheduler' not in kwargs.keys():
            kwargs['lr_scheduler'] = partial(
                ReduceLROnPlateau,
                factor=0.2,
                mode='max',
                patience=int(num_epoch / 10),
                verbose=True
            )

        fit_parameters = FitParameters(
            dataset_name=ds_name,
            train_dl=train_dl,
            val_dl=val_dl,
            num_epochs=num_epoch,
            models_path=os.path.join(self.path, 'models'),
            summary_path=os.path.join(self.path, 'summary'),
            **kwargs
        )
        if self.optim is None:
            self.exp.fit(p=fit_parameters)
        else:
            ft_num_epoch = ft_params.pop('num_epochs', 10)
            if 'lr_scheduler' not in ft_params.keys():
                ft_params['lr_scheduler'] = partial(
                    ReduceLROnPlateau,
                    factor=0.2,
                    mode='max',
                    patience=int(ft_num_epoch / 10),
                    verbose=True
                )
            ft_parameters = FitParameters(
                dataset_name=ds_name,
                train_dl=train_dl,
                val_dl=val_dl,
                num_epochs=ft_num_epoch,
                models_path=os.path.join(self.path, 'models'),
                summary_path=os.path.join(self.path, 'summary'),
                **ft_params
            )
            self.optim.fit(exp=self.exp, params=fit_parameters, ft_params=ft_parameters)

        return self.exp.model

    def predict(self, data_path, **kwargs):
        """
        Computes predictions for data in folder.

        Args:
            data_path: Path to image folder.

        Returns:
            Predictions dictionary.
        """
        dataset = PredictionFolderDataset(
            image_folder=data_path,
            transform=kwargs.pop('transform', ToTensor())
        )
        dataloader = DataLoader(
            dataset=dataset,
            collate_fn=(lambda x: tuple(zip(*x))) if self.task == 'object_detection' else None,
            **kwargs
        )
        predictions = self.exp.predict(dataloader=dataloader)
        return CV_TASKS[self.task]['idx_to_class'](predictions, self.idx_to_class)

    def predict_proba(self, data_path, **kwargs):
        """
        Computes probability predictions for data in folder.

        Args:
            data_path: Path to image folder.

        Returns:
            Predictions dictionary.
        """
        dataset = PredictionFolderDataset(
            image_folder=data_path,
            transform=kwargs.pop('transform', ToTensor())
        )
        dataloader = DataLoader(
            dataset=dataset,
            collate_fn=(lambda x: tuple(zip(*x))) if self.task == 'object_detection' else None,
            **kwargs
        )
        predictions = self.exp.predict_proba(dataloader=dataloader)
        return CV_TASKS[self.task]['idx_to_class'](predictions, self.idx_to_class, True)

    def get_metrics(self, **kwargs):
        """
        Runs model validation and returns metric values.

        Returns:
            Metrics dictionary
        """
        if self.val_dl is not None:
            return self.exp.val_loop(dataloader=self.val_dl, **kwargs)
        else:
            raise AttributeError('No validation data. Call the fit method before.')

    def load(self, path) -> None:
        """
        Load the model state dict.

        Args:
            path: Path to the model state dict.
        """
        if self.optim is None:
            self.exp.load_model(path)
        else:
            self.optim.load_model(self.exp, path)

    def save(self, path) -> None:
        """
        Save the model state dict.

        Args:
            path: Path to the model state dict.
        """
        self.exp.save_model(file_path=path)

    def save_metrics(self, **kwargs) -> None:
        """Displays an informational message with the metrics save path"""
        print(f'All metrics were saved during training in {self.path}')

    def save_predict(self, **kwargs) -> None:
        raise NotImplementedError



import numpy as np


class Anomaly:
    def __init__(self, params: dict):
        self.level = params.get('level', 10)
        self.anomaly_type = self.__class__.__name__

    def get(self, ts: np.ndarray, interval: tuple):
        NotImplementedError()


class ShiftTrendUP(Anomaly):
    def __init__(self, params):
        super().__init__(params)

    def get(self, ts: np.ndarray, interval: tuple):
        shift = np.zeros(ts.size)
        shift_value = np.mean(ts[interval[0]:interval[1] + 1]) * (self.level / 100)
        shift_value = abs(shift_value)
        shift[interval[0]:interval[1] + 1] = shift_value
        return self.apply_shift(ts, shift)

    def apply_shift(self, ts: np.ndarray, shift: np.ndarray):
        return ts + shift


class ShiftTrendDOWN(ShiftTrendUP):
    def __init__(self, params):
        super().__init__(params)

    def apply_shift(self, ts: np.ndarray, shift: np.ndarray):
        return ts - shift


class DecreaseDispersion(Anomaly):

    def __init__(self, params):
        super().__init__(params)

    def get(self, ts: np.ndarray, interval: tuple):
        new_ts = ts.copy()
        sector_values = new_ts[interval[0]:interval[1] + 1]
        mean = float(np.mean(sector_values))
        new_sector_values = [self.shrink(mean, x) for x in sector_values]
        new_ts[interval[0]:interval[1] + 1] = new_sector_values
        return new_ts

    def shrink(self, mean_value: float, i: float):
        diff = mean_value - i
        new_diff = diff - diff * (self.level / 100)
        new_i = mean_value - new_diff
        return new_i


class IncreaseDispersion(DecreaseDispersion):

    def __init__(self, params):
        super().__init__(params)

    def shrink(self, mean_value: float, i: float):
        diff = mean_value - i
        new_diff = diff + diff * (self.level / 100)
        new_i = mean_value - new_diff
        return new_i


class AddNoise(Anomaly):
    def __init__(self, params):
        super().__init__(params)
        self.noise_type = params.get('noise_type', np.random.choice(['gaussian',
                                                                     'uniform',
                                                                     'laplace']))

    def get(self, ts: np.ndarray, interval: tuple):
        ts_ = ts.copy()
        sector = ts_[interval[0]: interval[1]+1]

        noise_std = np.std(sector) * self.level / 100

        if self.noise_type == 'gaussian':
            noise = np.random.normal(0, noise_std, len(sector))
        elif self.noise_type == 'uniform':
            noise = np.random.uniform(-noise_std, noise_std, len(sector))
        elif self.noise_type == 'laplace':
            noise = np.random.laplace(0, noise_std, len(sector))
        else:
            raise ValueError("Invalid noise_type. Please choose 'gaussian', 'uniform', or 'laplace'.")

        noisy_sector = sector + noise
        ts_[interval[0]:interval[1]+1] = noisy_sector

        return ts_


class Peak(Anomaly):
    def __init__(self, params):
        super().__init__(params)

    def get(self, ts: np.ndarray, interval: tuple):
        ts_ = ts.copy()
        shift = np.zeros(ts.size)
        sector = ts_[interval[0]: interval[1]+1]
        peak_value = abs(np.mean(sector) * (self.level / 100))
        center_point = int((interval[1]+1 + interval[0]) / 2)
        shift[center_point] = peak_value
        return self.apply_shift(ts_, shift)

    def apply_shift(self, ts_, shift):
        return ts_ + shift


class Dip(Peak):
    def __init__(self, params):
        super().__init__(params)

    def apply_shift(self, ts_, shift):
        return ts_ - shift


class ChangeTrend(Anomaly):
    def __init__(self, params):
        super().__init__(params)

    def get(self, ts: np.ndarray, interval: tuple):
        pass



import logging
import os
import shutil
import urllib.request as request
import zipfile
from pathlib import Path

import chardet
import numpy as np
import pandas as pd
from scipy.io.arff import loadarff
from sktime.datasets._data_io import load_from_tsfile_to_dataframe
from tqdm import tqdm

from fedot_ind.api.utils.path_lib import PROJECT_PATH


class DataLoader:
    """Class for reading data from ``tsv`` files and downloading from UCR archive if not found locally.
    At the moment supports only ``.txt`` and ``.arff`` formats, but not relational ``.arff`` or ``.ts`` files.

    Args:
        dataset_name: name of dataset

    Examples:
        >>> data_loader = DataLoader('ItalyPowerDemand')
        >>> train_data, test_data = data_loader.load_data()
    """

    def __init__(self, dataset_name: str, folder: str = None):
        self.logger = logging.getLogger('DataLoader')
        self.dataset_name = dataset_name
        self.folder = folder

    def load_data(self) -> tuple:
        """Load data for classification experiment locally or externally from UCR archive.

        Returns:
            tuple: train and test data
        """
        dataset_name = self.dataset_name

        data_path = os.path.join(PROJECT_PATH, 'fedot_ind', 'data') if self.folder is None else self.folder

        _, train_data, test_data = self.read_train_test_files(dataset_name=dataset_name,
                                                              data_path=data_path)

        if train_data is None:
            self.logger.info(f'Downloading...')

            # Create temporary folder for downloaded data
            cache_path = os.path.join(PROJECT_PATH, 'temp_cache/')
            download_path = cache_path + 'downloads/'
            temp_data_path = cache_path + 'temp_data/'
            filename = 'temp_data_{}'.format(dataset_name)
            for _ in (download_path, temp_data_path):
                os.makedirs(_, exist_ok=True)

            url = f"http://www.timeseriesclassification.com/aeon-toolkit/{dataset_name}.zip"
            request.urlretrieve(url, download_path + filename)
            try:
                zipfile.ZipFile(download_path + filename).extractall(temp_data_path + dataset_name)
            except zipfile.BadZipFile:
                self.logger.error(f'Cannot extract data: {dataset_name} dataset not found in UCR archive')
                return None, None

            self.logger.info(f'{dataset_name} data downloaded. Unpacking...')
            train_data, test_data = self.extract_data(dataset_name, temp_data_path)

            shutil.rmtree(cache_path)
            return train_data, test_data
        self.logger.info(f'Data readed successfully from local folder')
        return train_data, test_data

    def read_train_test_files(self, data_path, dataset_name):

        file_path = data_path + '/' + dataset_name + f'/{dataset_name}_TRAIN'
        # If data unpacked as .tsv file
        if os.path.isfile(file_path + '.tsv'):
            self.logger.info(f'Reading data from {data_path + "/" + dataset_name}')
            x_train, y_train, x_test, y_test = self.read_tsv(dataset_name, data_path)
            is_multi = False

        # If data unpacked as .txt file
        elif os.path.isfile(file_path + '.txt'):
            self.logger.info(f'Reading data from {data_path + "/" + dataset_name}')
            x_train, y_train, x_test, y_test = self.read_txt_files(dataset_name, data_path)
            is_multi = False

        # If data unpacked as .ts file
        elif os.path.isfile(file_path + '.ts'):
            self.logger.info(f'Reading data from {data_path + "/" + dataset_name}')
            x_train, y_train, x_test, y_test = self.read_ts_files(dataset_name, data_path)
            is_multi = True

        # If data unpacked as .arff file
        elif os.path.isfile(file_path + '.arff'):
            self.logger.info(f'Reading data from {data_path + "/" + dataset_name}')
            x_train, y_train, x_test, y_test = self.read_arff_files(dataset_name, data_path)
            is_multi = True

        else:
            self.logger.error(f'Data not found in {data_path + "/" + dataset_name}')
            return None, None, None
        return is_multi, (x_train, y_train), (x_test, y_test)

    def predict_encoding(self, file_path: Path, n_lines: int = 20) -> str:
        with Path(file_path).open('rb') as f:
            rawdata = b''.join([f.readline() for _ in range(n_lines)])
        return chardet.detect(rawdata)['encoding']

    def _load_from_tsfile_to_dataframe(self, full_file_path_and_name, return_separate_X_and_y=True,
                                       replace_missing_vals_with='NaN'):
        """Loads data from a .ts file into a Pandas DataFrame.
        Taken from https://github.com/ChangWeiTan/TS-Extrinsic-Regression/blob/master/utils/data_loader.py

        Args: full_file_path_and_name: The full pathname of the .ts file to read. return_separate_X_and_y: true if X
              and Y values should be returned as separate Data Frames (X) and a numpy array (y), false otherwise.
              replace_missing_vals_with: The value that missing values in the text file should be replaced with prior to
              parsing.

        Returns: DataFrame, ndarray: If return_separate_X_and_y then a tuple containing a DataFrame and a numpy array
                 containing the relevant time-series and corresponding class values.
                 DataFrame: If not return_separate_X_and_y then a single DataFrame containing all time-series and
                 (if relevant) a column "class_vals" the associated class values.

        """

        # Initialize flags and variables used when parsing the file
        metadata_started = False
        data_started = False

        has_problem_name_tag = False
        has_timestamps_tag = False
        has_univariate_tag = False
        has_class_labels_tag = False
        has_target_labels_tag = False
        has_data_tag = False

        previous_timestamp_was_float = None
        previous_timestamp_was_int = None
        previous_timestamp_was_timestamp = None
        num_dimensions = None
        is_first_case = True
        instance_list = []
        class_val_list = []
        line_num = 0
        TsFileParseException = Exception

        encoding = self.predict_encoding(full_file_path_and_name)

        with open(full_file_path_and_name, 'r', encoding=encoding) as file:
            dataset_name = os.path.basename(full_file_path_and_name)
            for line in tqdm(file.readlines(), desc='Loading data', leave=False, postfix=dataset_name, unit='lines'):
                # print(".", end='')
                # Strip white space from start/end of line and change to lowercase for use below
                line = line.strip().lower()
                # Empty lines are valid at any point in a file
                if line:
                    # Check if this line contains metadata
                    # Please note that even though metadata is stored in this function it is not currently published externally
                    if line.startswith("@problemname"):
                        # Check that the data has not started
                        if data_started:
                            raise TsFileParseException("metadata must come before data")
                        # Check that the associated value is valid
                        tokens = line.transform_for_fit(' ')
                        token_len = len(tokens)

                        if token_len == 1:
                            raise TsFileParseException("problemname tag requires an associated value")

                        problem_name = line[len("@problemname") + 1:]
                        has_problem_name_tag = True
                        metadata_started = True
                    elif line.startswith("@timestamps"):
                        # Check that the data has not started
                        if data_started:
                            raise TsFileParseException("metadata must come before data")

                        # Check that the associated value is valid
                        tokens = line.transform_for_fit(' ')
                        token_len = len(tokens)

                        if token_len != 2:
                            raise TsFileParseException("timestamps tag requires an associated Boolean value")
                        elif tokens[1] == "true":
                            timestamps = True
                        elif tokens[1] == "false":
                            timestamps = False
                        else:
                            raise TsFileParseException("invalid timestamps value")
                        has_timestamps_tag = True
                        metadata_started = True
                    elif line.startswith("@univariate"):
                        # Check that the data has not started
                        if data_started:
                            raise TsFileParseException("metadata must come before data")

                        # Check that the associated value is valid
                        tokens = line.transform_for_fit(' ')
                        token_len = len(tokens)
                        if token_len != 2:
                            raise TsFileParseException("univariate tag requires an associated Boolean value")
                        elif tokens[1] == "true":
                            univariate = True
                        elif tokens[1] == "false":
                            univariate = False
                        else:
                            raise TsFileParseException("invalid univariate value")

                        has_univariate_tag = True
                        metadata_started = True
                    elif line.startswith("@classlabel"):
                        # Check that the data has not started
                        if data_started:
                            raise TsFileParseException("metadata must come before data")

                        # Check that the associated value is valid
                        tokens = line.transform_for_fit(' ')
                        token_len = len(tokens)

                        if token_len == 1:
                            raise TsFileParseException("classlabel tag requires an associated Boolean value")

                        if tokens[1] == "true":
                            class_labels = True
                        elif tokens[1] == "false":
                            class_labels = False
                        else:
                            raise TsFileParseException("invalid classLabel value")

                        # Check if we have any associated class values
                        if token_len == 2 and class_labels:
                            raise TsFileParseException(
                                "if the classlabel tag is true then class values must be supplied")

                        has_class_labels_tag = True
                        class_label_list = [token.strip() for token in tokens[2:]]
                        metadata_started = True
                    elif line.startswith("@targetlabel"):
                        # Check that the data has not started
                        if data_started:
                            raise TsFileParseException("metadata must come before data")

                        # Check that the associated value is valid
                        tokens = line.transform_for_fit(' ')
                        token_len = len(tokens)

                        if token_len == 1:
                            raise TsFileParseException("targetlabel tag requires an associated Boolean value")

                        if tokens[1] == "true":
                            target_labels = True
                        elif tokens[1] == "false":
                            target_labels = False
                        else:
                            raise TsFileParseException("invalid targetLabel value")

                        has_target_labels_tag = True
                        class_val_list = []
                        metadata_started = True
                    # Check if this line contains the start of data
                    elif line.startswith("@data"):
                        if line != "@data":
                            raise TsFileParseException("data tag should not have an associated value")

                        if data_started and not metadata_started:
                            raise TsFileParseException("metadata must come before data")
                        else:
                            has_data_tag = True
                            data_started = True
                    # If the 'data tag has been found then metadata has been parsed and data can be loaded
                    elif data_started:
                        # Check that a full set of metadata has been provided
                        incomplete_regression_meta_data = not has_problem_name_tag or not has_timestamps_tag or not has_univariate_tag or not has_target_labels_tag or not has_data_tag
                        incomplete_classification_meta_data = not has_problem_name_tag or not has_timestamps_tag or not has_univariate_tag or not has_class_labels_tag or not has_data_tag
                        if incomplete_regression_meta_data and incomplete_classification_meta_data:
                            raise TsFileParseException("a full set of metadata has not been provided before the data")

                        # Replace any missing values with the value specified
                        line = line.replace("?", replace_missing_vals_with)

                        # Check if we dealing with data that has timestamps
                        if timestamps:
                            # We're dealing with timestamps so cannot just split line on ':' as timestamps may contain one
                            has_another_value = False
                            has_another_dimension = False

                            timestamps_for_dimension = []
                            values_for_dimension = []

                            this_line_num_dimensions = 0
                            line_len = len(line)
                            char_num = 0

                            while char_num < line_len:
                                # Move through any spaces
                                while char_num < line_len and str.isspace(line[char_num]):
                                    char_num += 1

                                # See if there is any more data to read in or if we should validate that read thus far

                                if char_num < line_len:

                                    # See if we have an empty dimension (i.e. no values)
                                    if line[char_num] == ":":
                                        if len(instance_list) < (this_line_num_dimensions + 1):
                                            instance_list.append([])

                                        instance_list[this_line_num_dimensions].append(pd.Series())
                                        this_line_num_dimensions += 1

                                        has_another_value = False
                                        has_another_dimension = True

                                        timestamps_for_dimension = []
                                        values_for_dimension = []

                                        char_num += 1
                                    else:
                                        # Check if we have reached a class label
                                        if line[char_num] != "(" and target_labels:
                                            class_val = line[char_num:].strip()

                                            # if class_val not in class_val_list:
                                            #     raise TsFileParseException(
                                            #         "the class value '" + class_val + "' on line " + str(
                                            #             line_num + 1) + " is not valid")

                                            class_val_list.append(float(class_val))
                                            char_num = line_len

                                            has_another_value = False
                                            has_another_dimension = False

                                            timestamps_for_dimension = []
                                            values_for_dimension = []

                                        else:

                                            # Read in the data contained within the next tuple

                                            if line[char_num] != "(" and not target_labels:
                                                raise TsFileParseException(
                                                    "dimension " + str(
                                                        this_line_num_dimensions + 1) + " on line " + str(
                                                        line_num + 1) + " does not start with a '('")

                                            char_num += 1
                                            tuple_data = ""

                                            while char_num < line_len and line[char_num] != ")":
                                                tuple_data += line[char_num]
                                                char_num += 1

                                            if char_num >= line_len or line[char_num] != ")":
                                                raise TsFileParseException(
                                                    "dimension " + str(
                                                        this_line_num_dimensions + 1) + " on line " + str(
                                                        line_num + 1) + " does not end with a ')'")

                                            # Read in any spaces immediately after the current tuple

                                            char_num += 1

                                            while char_num < line_len and str.isspace(line[char_num]):
                                                char_num += 1

                                            # Check if there is another value or dimension to process after this tuple

                                            if char_num >= line_len:
                                                has_another_value = False
                                                has_another_dimension = False

                                            elif line[char_num] == ",":
                                                has_another_value = True
                                                has_another_dimension = False

                                            elif line[char_num] == ":":
                                                has_another_value = False
                                                has_another_dimension = True

                                            char_num += 1

                                            # Get the numeric value for the tuple by reading from the end of the tuple data backwards to the last comma

                                            last_comma_index = tuple_data.rfind(',')

                                            if last_comma_index == -1:
                                                raise TsFileParseException(
                                                    "dimension " + str(
                                                        this_line_num_dimensions + 1) + " on line " + str(
                                                        line_num + 1) + " contains a tuple that has no comma inside of it")

                                            try:
                                                value = tuple_data[last_comma_index + 1:]
                                                value = float(value)

                                            except ValueError:
                                                raise TsFileParseException(
                                                    "dimension " + str(
                                                        this_line_num_dimensions + 1) + " on line " + str(
                                                        line_num + 1) + " contains a tuple that does not have a valid numeric value")

                                            # Check the type of timestamp that we have

                                            timestamp = tuple_data[0: last_comma_index]

                                            try:
                                                timestamp = int(timestamp)
                                                timestamp_is_int = True
                                                timestamp_is_timestamp = False
                                            except ValueError:
                                                timestamp_is_int = False

                                            if not timestamp_is_int:
                                                try:
                                                    timestamp = float(timestamp)
                                                    timestamp_is_float = True
                                                    timestamp_is_timestamp = False
                                                except ValueError:
                                                    timestamp_is_float = False

                                            if not timestamp_is_int and not timestamp_is_float:
                                                try:
                                                    timestamp = timestamp.strip()
                                                    timestamp_is_timestamp = True
                                                except ValueError:
                                                    timestamp_is_timestamp = False

                                            # Make sure that the timestamps in the file (not just this dimension or case) are consistent

                                            if not timestamp_is_timestamp and not timestamp_is_int and not timestamp_is_float:
                                                raise TsFileParseException(
                                                    "dimension " + str(
                                                        this_line_num_dimensions + 1) + " on line " + str(
                                                        line_num + 1) + " contains a tuple that has an invalid timestamp '" + timestamp + "'")

                                            if previous_timestamp_was_float is not None and previous_timestamp_was_float and not timestamp_is_float:
                                                raise TsFileParseException(
                                                    "dimension " + str(
                                                        this_line_num_dimensions + 1) + " on line " + str(
                                                        line_num + 1) + " contains tuples where the timestamp format is inconsistent")

                                            if previous_timestamp_was_int is not None and previous_timestamp_was_int and not timestamp_is_int:
                                                raise TsFileParseException(
                                                    "dimension " + str(
                                                        this_line_num_dimensions + 1) + " on line " + str(
                                                        line_num + 1) + " contains tuples where the timestamp format is inconsistent")

                                            if previous_timestamp_was_timestamp is not None and previous_timestamp_was_timestamp and not timestamp_is_timestamp:
                                                raise TsFileParseException(
                                                    "dimension " + str(
                                                        this_line_num_dimensions + 1) + " on line " + str(
                                                        line_num + 1) + " contains tuples where the timestamp format is inconsistent")

                                            # Store the values

                                            timestamps_for_dimension += [timestamp]
                                            values_for_dimension += [value]

                                            #  If this was our first tuple then we store the type of timestamp we had

                                            if previous_timestamp_was_timestamp is None and timestamp_is_timestamp:
                                                previous_timestamp_was_timestamp = True
                                                previous_timestamp_was_int = False
                                                previous_timestamp_was_float = False

                                            if previous_timestamp_was_int is None and timestamp_is_int:
                                                previous_timestamp_was_timestamp = False
                                                previous_timestamp_was_int = True
                                                previous_timestamp_was_float = False

                                            if previous_timestamp_was_float is None and timestamp_is_float:
                                                previous_timestamp_was_timestamp = False
                                                previous_timestamp_was_int = False
                                                previous_timestamp_was_float = True

                                            # See if we should add the data for this dimension

                                            if not has_another_value:
                                                if len(instance_list) < (this_line_num_dimensions + 1):
                                                    instance_list.append([])

                                                if timestamp_is_timestamp:
                                                    timestamps_for_dimension = pd.DatetimeIndex(
                                                        timestamps_for_dimension)

                                                instance_list[this_line_num_dimensions].append(
                                                    pd.Series(index=timestamps_for_dimension,
                                                              data=values_for_dimension))
                                                this_line_num_dimensions += 1

                                                timestamps_for_dimension = []
                                                values_for_dimension = []

                                elif has_another_value:
                                    raise TsFileParseException(
                                        "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                            line_num + 1) + " ends with a ',' that is not followed by another tuple")

                                elif has_another_dimension and target_labels:
                                    raise TsFileParseException(
                                        "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                            line_num + 1) + " ends with a ':' while it should list a class value")

                                elif has_another_dimension and not target_labels:
                                    if len(instance_list) < (this_line_num_dimensions + 1):
                                        instance_list.append([])

                                    instance_list[this_line_num_dimensions].append(pd.Series(dtype=np.float32))
                                    this_line_num_dimensions += 1
                                    num_dimensions = this_line_num_dimensions

                                # If this is the 1st line of data we have seen then note the dimensions

                                if not has_another_value and not has_another_dimension:
                                    if num_dimensions is None:
                                        num_dimensions = this_line_num_dimensions

                                    if num_dimensions != this_line_num_dimensions:
                                        raise TsFileParseException("line " + str(
                                            line_num + 1) + " does not have the same number of dimensions as the previous line of data")

                            # Check that we are not expecting some more data, and if not, store that processed above

                            if has_another_value:
                                raise TsFileParseException(
                                    "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                        line_num + 1) + " ends with a ',' that is not followed by another tuple")

                            elif has_another_dimension and target_labels:
                                raise TsFileParseException(
                                    "dimension " + str(this_line_num_dimensions + 1) + " on line " + str(
                                        line_num + 1) + " ends with a ':' while it should list a class value")

                            elif has_another_dimension and not target_labels:
                                if len(instance_list) < (this_line_num_dimensions + 1):
                                    instance_list.append([])

                                instance_list[this_line_num_dimensions].append(pd.Series())
                                this_line_num_dimensions += 1
                                num_dimensions = this_line_num_dimensions

                            # If this is the 1st line of data we have seen then note the dimensions

                            if not has_another_value and num_dimensions != this_line_num_dimensions:
                                raise TsFileParseException("line " + str(
                                    line_num + 1) + " does not have the same number of dimensions as the previous line of data")

                            # Check if we should have class values, and if so that they are contained in those listed in the metadata

                            if target_labels and len(class_val_list) == 0:
                                raise TsFileParseException("the cases have no associated class values")
                        else:
                            dimensions = line.transform_for_fit(":")
                            # If first row then note the number of dimensions (that must be the same for all cases)
                            if is_first_case:
                                num_dimensions = len(dimensions)

                                if target_labels:
                                    num_dimensions -= 1

                                for dim in range(0, num_dimensions):
                                    instance_list.append([])
                                is_first_case = False

                            # See how many dimensions that the case whose data in represented in this line has
                            this_line_num_dimensions = len(dimensions)

                            if target_labels:
                                this_line_num_dimensions -= 1

                            # All dimensions should be included for all series, even if they are empty
                            if this_line_num_dimensions != num_dimensions:
                                raise TsFileParseException("inconsistent number of dimensions. Expecting " + str(
                                    num_dimensions) + " but have read " + str(this_line_num_dimensions))

                            # Process the data for each dimension
                            for dim in range(0, num_dimensions):
                                dimension = dimensions[dim].strip()

                                if dimension:
                                    data_series = dimension.transform_for_fit(",")
                                    data_series = [float(i) for i in data_series]
                                    instance_list[dim].append(pd.Series(data_series))
                                else:
                                    instance_list[dim].append(pd.Series())

                            if target_labels:
                                class_val_list.append(float(dimensions[num_dimensions].strip()))

                line_num += 1

        # Check that the file was not empty
        if line_num:
            # Check that the file contained both metadata and data
            complete_regression_meta_data = has_problem_name_tag and has_timestamps_tag and has_univariate_tag and has_target_labels_tag and has_data_tag
            complete_classification_meta_data = has_problem_name_tag and has_timestamps_tag and has_univariate_tag and has_class_labels_tag and has_data_tag

            if metadata_started and not complete_regression_meta_data and not complete_classification_meta_data:
                raise TsFileParseException("metadata incomplete")
            elif metadata_started and not data_started:
                raise TsFileParseException("file contained metadata but no data")
            elif metadata_started and data_started and len(instance_list) == 0:
                raise TsFileParseException("file contained metadata but no data")

            # Create a DataFrame from the data parsed above
            data = pd.DataFrame(dtype=np.float32)

            for dim in range(0, num_dimensions):
                data['dim_' + str(dim)] = instance_list[dim]

            # Check if we should return any associated class labels separately

            if target_labels:
                if return_separate_X_and_y:
                    return data, np.asarray(class_val_list)
                else:
                    data['class_vals'] = pd.Series(class_val_list)
                    return data
            else:
                return data
        else:
            raise TsFileParseException("empty file")

    def read_tsv(self, dataset_name: str, data_path: str) -> tuple:
        """Read ``tsv`` file that contains data for classification experiment. Data must be placed
        in ``data`` folder with ``.tsv`` extension.

        Args:
            dataset_name: name of dataset
            data_path: path to temporary folder with downloaded data
        Returns:
            tuple: (x_train, x_test) and (y_train, y_test)

        """
        df_train = pd.read_csv(os.path.join(data_path, dataset_name, f'{dataset_name}_TRAIN.tsv'),
                               sep='\t',
                               header=None)

        x_train = df_train.iloc[:, 1:]
        y_train = df_train[0].values

        df_test = pd.read_csv(os.path.join(data_path, dataset_name, f'{dataset_name}_TEST.tsv'),
                              sep='\t',
                              header=None)

        x_test = df_test.iloc[:, 1:]
        y_test = df_test[0].values
        try:
            y_train, y_test = y_train.astype(int), y_test.astype(int)
        except ValueError:
            y_train, y_test = y_train.astype(str), y_test.astype(str)

        return x_train, y_train, x_test, y_test

    @staticmethod
    def read_txt_files(dataset_name: str, temp_data_path: str):
        """
        Reads data from ``.txt`` file.

        Args:
            dataset_name: name of dataset
            temp_data_path: path to temporary folder with downloaded data

        Returns:
            train and test data tuple
        """
        data_train = np.genfromtxt(temp_data_path + '/' + dataset_name + f'/{dataset_name}_TRAIN.txt')
        data_test = np.genfromtxt(temp_data_path + '/' + dataset_name + f'/{dataset_name}_TEST.txt')
        x_train, y_train = data_train[:, 1:], data_train[:, 0]
        x_test, y_test = data_test[:, 1:], data_test[:, 0]
        return x_train, y_train, x_test, y_test

    def read_ts_files(self, dataset_name, data_path):
        try:
            x_test, y_test = load_from_tsfile_to_dataframe(data_path + '/' + dataset_name + f'/{dataset_name}_TEST.ts',
                                                           return_separate_X_and_y=True)
            x_train, y_train = load_from_tsfile_to_dataframe(
                data_path + '/' + dataset_name + f'/{dataset_name}_TRAIN.ts',
                return_separate_X_and_y=True)
            return x_train, y_train, x_test, y_test
        except Exception:
            x_test, y_test = self._load_from_tsfile_to_dataframe(
                data_path + '/' + dataset_name + f'/{dataset_name}_TEST.ts',
                return_separate_X_and_y=True)
            x_train, y_train = self._load_from_tsfile_to_dataframe(
                data_path + '/' + dataset_name + f'/{dataset_name}_TRAIN.ts',
                return_separate_X_and_y=True)
            return x_train, y_train, x_test, y_test

    def read_arff_files(self, dataset_name, temp_data_path):
        """Reads data from ``.arff`` file.

        """
        train = loadarff(temp_data_path + dataset_name + f'/{dataset_name}_TRAIN.arff')
        test = loadarff(temp_data_path + dataset_name + f'/{dataset_name}_TEST.arff')

        data_train = np.asarray([train[0][name] for name in train[1].names()])
        x_train = data_train[:-1].T.astype('float64')
        y_train = data_train[-1]

        data_test = np.asarray([test[0][name] for name in test[1].names()])
        x_test = data_test[:-1].T.astype('float64')
        y_test = data_test[-1]
        return x_train, y_train, x_test, y_test

    def extract_data(self, dataset_name: str, data_path: str):
        """Unpacks data from downloaded file and saves it into Data folder with ``.tsv`` extension.

        Args:
            dataset_name: name of dataset
            data_path: path to folder downloaded data

        Returns:
            tuple: train and test data

        """
        try:
            is_multi, (x_train, y_train), (x_test, y_test) = self.read_train_test_files(data_path, dataset_name)

        except Exception as e:
            self.logger.error(f'Error while unpacking data: {e}')
            return None, None

        # Conversion of target values to int or str
        try:
            y_train = y_train.astype('int64')
            y_test = y_test.astype('int64')
        except ValueError:
            y_train = y_train.astype(str)
            y_test = y_test.astype(str)

        # Save data to tsv files
        new_path = os.path.join(PROJECT_PATH, 'fedot_ind','data', dataset_name)
        os.makedirs(new_path, exist_ok=True)

        self.logger.info(f'Saving {dataset_name} data files')
        for subset in ('TRAIN', 'TEST'):
            if not is_multi:
                df = pd.DataFrame(x_train if subset == 'TRAIN' else x_test)
                df.insert(0, 'class', y_train if subset == 'TRAIN' else y_test)
                df.to_csv(os.path.join(new_path, f'{dataset_name}_{subset}.tsv'), sep='\t', index=False, header=False)
                del df

            else:
                old_path = os.path.join(data_path, dataset_name, f'{dataset_name}_{subset}.ts')
                shutil.move(old_path, new_path)

        if is_multi:
            return (x_train, y_train), (x_test, y_test)
        else:
            return (pd.DataFrame(x_train), y_train), (pd.DataFrame(x_test), y_test)


if __name__ == '__main__':
    data_loader = DataLoader('Car')
    _train_data, _test_data = data_loader.load_data()



"""This module contains classes for working with neural networks using pytorch."""
import logging
import os
import shutil
from functools import partial
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List, Optional, Type, Union
from abc import ABC, abstractmethod

import torch
from torch.nn.functional import softmax
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import ReduceLROnPlateau, LRScheduler
from tqdm import tqdm

from fedot_ind.core.architecture.abstraction.writers import CSVWriter, TFWriter, WriterComposer
from fedot_ind.core.architecture.abstraction.сheckers import parameter_value_check
from fedot_ind.core.metrics.cv_metrics import MetricCounter, ClassificationMetricCounter, LossesAverager, \
    ObjectDetectionMetricCounter, SegmentationMetricCounter


@dataclass(frozen=True)
class FitParameters:
    """The data class containing the training parameters.

    Args:
        dataset_name: Name of dataset.
        train_dl: Train dataloader.
        val_dl: Validation dataloader.
        num_epochs: Number of training epochs.
        optimizer: Type of model optimizer, e.g. ``torch.optim.Adam``.
        lr_scheduler: Type of learning rate scheduler, e.g ``torch.optim.lr_scheduler.StepLR``.
        models_path: Path to folder for saving models.
        summary_path: Path to folder for writing experiment summary info.
        validation_period: Validation frequency in epochs.
        class_metrics: If ``True``, calculates validation metrics for each class.
        description: Additional line describing the experiment.

    """

    dataset_name: str
    train_dl: DataLoader
    val_dl: DataLoader
    num_epochs: int
    optimizer: Union[Type[torch.optim.Optimizer], partial] = torch.optim.Adam
    lr_scheduler: Optional[Union[Type[torch.optim.lr_scheduler.LRScheduler], partial]] = None
    models_path: Union[Path, str] = 'models'
    summary_path: Union[Path, str] = 'summary'
    validation_period: int = 1
    class_metrics: bool = False
    description: str = ''


class NNExperimenter(ABC):
    """Generalized class for working with neural models.

    Args:
        model: Trainable model.
        metric: Target metric by which models are compared.
        metric_counter: Class for calculating metrics.
            Must implement ``update`` and ``compute`` methods.
        name: Name of the model.
        weights: Path to the model state_dict to load weights.
        device: String passed to ``torch.device`` initialization.

    """

    def __init__(
            self,
            model: torch.nn.Module,
            metric: str,
            metric_counter: Type[MetricCounter],
            name: Optional[str],
            weights: Optional[str],
            device: str,
    ):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.model = model
        self.device = torch.device(device)
        if weights is not None:
            self.model.load_state_dict(torch.load(weights, map_location=self.device))
        self.model.to(self.device)
        self.name = name if name is not None else type(model).__name__
        self.best_score = -1
        self.metric = metric
        self.metric_counter = metric_counter

    def fit(self,
            p: FitParameters,
            phase: str = 'train',
            model_losses: Optional[Callable] = None,
            filter_pruning: Optional[Dict] = None,
            start_epoch: int = 0,
            initial_validation: bool = False
            ) -> None:
        """Run model training.

        Args:
            p: An object containing training parameters.
            phase: String explanation of training.
            model_losses: Function for calculating losses from model weights.
            filter_pruning: Parameters (pruning function and condition) passed to ``apply_func`` function.
            start_epoch: Initial training epoch.
            initial_validation: If ``True`` run validation loop before training.

        """
        model_path = os.path.join(p.models_path, p.dataset_name, self.name, p.description, phase)
        summary_path = os.path.join(p.summary_path, p.dataset_name, self.name, p.description, phase)
        writer = WriterComposer(summary_path, [TFWriter, CSVWriter])
        self.logger.info(f"{phase}: {self.name}, using device: {self.device}")

        if initial_validation:
            init_scores = self.val_loop(dataloader=p.val_dl, class_metrics=p.class_metrics)
            writer.write_scores('val', init_scores, start_epoch)
            self._save_model_sd_if_best(val_scores=init_scores, file_path=model_path)
        start_epoch += 1

        optimizer = p.optimizer(self.model.parameters())
        lr_scheduler = None
        if p.lr_scheduler is not None:
            lr_scheduler = p.lr_scheduler(optimizer)

        for epoch in range(start_epoch, start_epoch + p.num_epochs):
            self.logger.info(f"Epoch {epoch}")
            train_scores = self.train_loop(
                dataloader=p.train_dl,
                optimizer=optimizer,
                model_losses=model_losses
            )
            writer.write_scores('train', train_scores, epoch)

            if filter_pruning is not None:
                self._apply_function(**filter_pruning)

            if epoch % p.validation_period == 0:
                val_scores = self.val_loop(
                    dataloader=p.val_dl,
                    class_metrics=p.class_metrics
                )
                writer.write_scores('val', val_scores, epoch)
                self._save_model_sd_if_best(val_scores=val_scores, file_path=model_path)
                if isinstance(lr_scheduler, ReduceLROnPlateau):
                    lr_scheduler.step(val_scores[self.metric])

            if isinstance(lr_scheduler, LRScheduler) and not isinstance(lr_scheduler, ReduceLROnPlateau):
                lr_scheduler.step()
        self.load_model(model_path)
        self.logger.info(f'{self.metric} score: {self.best_score}')
        writer.close()

    def _save_model_sd_if_best(self, val_scores: Dict, file_path):
        """Save the model state dict if the best result on the target metric is achieved.

        Args:
            val_scores: Validation metric dictionary.
            file_path: Path to the file without extension.

        """
        if val_scores[self.metric] > self.best_score:
            self.best_score = val_scores[self.metric]
            self.logger.info(f'Best {self.metric} score: {self.best_score}')
            self.save_model(file_path=file_path)

    def save_model(
            self,
            file_path: str,
            state_dict: bool = True,
    ) -> None:
        """Save the model or its state dict.

        Args:
            file_path: Path to the file without extension.
            state_dict: If ``True`` save state_dict with extension ".sd.pt",
                else save all model with extension ".model.pt".

        """
        dir_path, file_name = os.path.split(file_path)
        os.makedirs(dir_path, exist_ok=True)
        file_name = f"{file_name}.{'sd' if state_dict else 'model'}.pt"
        file_path = os.path.join(dir_path, file_name)
        data = self.model.state_dict() if state_dict else self.model
        try:
            torch.save(data, file_path)
        except Exception:
            torch.save(data, file_name)
            shutil.move(file_name, dir_path)
        self.logger.info(f"Saved to {os.path.abspath(file_path)}.")

    def load_model(
            self,
            file_path: str,
            state_dict: bool = True,
    ) -> None:
        """Load the model or its state dict.

        Args:
            file_path: Path to the file without extension.
            state_dict: If ``True`` load state_dict with extension ".sd.pt",
                else load all model with extension ".model.pt".
        """
        file_path = f"{file_path}.{'sd' if state_dict else 'model'}.pt"
        data = torch.load(file_path, map_location=self.device)
        if state_dict:
            self.model.load_state_dict(data)
            self.model.to(self.device)
            self.logger.info("Model state dict loaded.")
        else:
            self.model = data
            self.logger.info("Model loaded.")

    def size_of_model(self) -> float:
        """Returns size of model in Mb."""
        param_size = 0
        for param in self.model.parameters():
            param_size += param.nelement() * param.element_size()
        buffer_size = 0
        for buffer in self.model.buffers():
            buffer_size += buffer.nelement() * buffer.element_size()
        return (param_size + buffer_size) / 1e6

    def number_of_model_params(self) -> int:
        """Returns number of model parameters."""
        return sum(p.numel() for p in self.model.parameters())

    def _apply_function(
            self,
            func: Callable,
            condition: Optional[Callable] = None
    ):
        """Applies the passed function to model layers by condition.

        Args:
            func: Applicable function.
            condition: Condition function for layer filtering.
        """
        for module in filter(condition, self.model.modules()):
            func(module)

    @abstractmethod
    def _forward(self, x: torch.Tensor):
        """Have to implement the forward method of the model and return predictions."""
        raise NotImplementedError

    @abstractmethod
    def _forward_with_loss(self, x: torch.Tensor, y) -> Dict[str, torch.Tensor]:
        """Have to implement the train forward method and return dictionary of losses."""
        raise NotImplementedError

    @abstractmethod
    def _predict_on_batch(self, x: torch.Tensor, proba: bool) -> List:
        """Have to implement the prediction method on batch."""
        raise NotImplementedError

    def predict(
            self,
            dataloader: DataLoader,
            proba: bool = False,
    ) -> Dict:
        """Computes predictions for data in dataloader.

        Args:
            dataloader: Data loader with prediction dataset.
            proba: If ``True`` computes probabilities.
        """
        self.logger.info('Computing predictions')
        ids = []
        preds = []
        self.model.eval()
        with torch.no_grad():
            for x, id in tqdm(dataloader, desc='predict'):
                ids.extend(id)
                preds.extend(self._predict_on_batch(x, proba=proba))
        return dict(zip(ids, preds))

    def predict_proba(self, dataloader: DataLoader) -> Dict:
        """Computes probability predictions for data in dataloader.

        Args:
            dataloader: Data loader with prediction dataset.
        """
        return self.predict(dataloader, proba=True)

    def train_loop(
            self,
            dataloader: torch.utils.data.DataLoader,
            optimizer: torch.optim.Optimizer,
            model_losses: Optional[Callable] = None,
    ) -> Dict[str, float]:
        """Training method of the model.

        Args:
            dataloader: Training data loader.
            optimizer: Model optimizer.
            model_losses: Function for calculating losses from model weights.

        Returns:
            Dict: {metric_name: value}.
        """
        self.model.train()
        train_scores = LossesAverager()
        batches = tqdm(dataloader, desc='train')
        for x, y in batches:
            losses = self._forward_with_loss(x, y)
            if model_losses is not None:
                losses.update(model_losses(self.model))
            train_scores.update(losses)
            loss = sum(losses.values())
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            batches.set_postfix(train_scores.compute())
        return train_scores.compute()

    def val_loop(
            self,
            dataloader: torch.utils.data.DataLoader,
            class_metrics: bool = False,
    ) -> Dict[str, float]:
        """Validation method of the model. Returns val_scores

        Args:
            dataloader: Validation data loader.
            class_metrics: If ``True``, calculates validation metrics for each class.

        Returns:
            Dict: {metric_name: value}.
        """
        self.model.eval()
        metric = self.metric_counter(class_metrics=class_metrics)
        with torch.no_grad():
            for x, y in tqdm(dataloader, desc='val'):
                preds = self._forward(x)
                metric.update(preds, y)
        return metric.compute()


class ClassificationExperimenter(NNExperimenter):
    """Class for working with classification models.

    Args:
        model: Trainable model.
        metric: Target metric by which models are compared.
            One of ``'f1'``, ``'accuracy'``, ``'precision'``, ``'recall'``, ``'roc_auc'``.
        loss: Loss function applied to model output.
        name: Name of the model.
        weights: Path to the model state_dict to load weights.
        device: String passed to ``torch.device`` initialization.
    """

    def __init__(
            self,
            model: torch.nn.Module,
            metric: str = 'f1',
            loss: Callable = torch.nn.CrossEntropyLoss(),
            name: Optional[str] = None,
            weights: Optional[str] = None,
            device: str = 'cuda',
    ):
        parameter_value_check(
            parameter='metric',
            value=metric,
            valid_values={'f1', 'accuracy', 'precision', 'recall', 'roc_auc'},
        )
        super().__init__(
            model=model,
            metric=metric,
            metric_counter=ClassificationMetricCounter,
            name=name,
            weights=weights,
            device=device
        )
        self.loss = loss

    def _forward(self, x: torch.Tensor) -> torch.Tensor:
        """Implements the forward method of the model and returns predictions."""
        x = x.to(self.device)
        return self.model(x)

    def _forward_with_loss(self, x: torch.Tensor, y) -> Dict[str, torch.Tensor]:
        """Implements the train forward method and returns loss."""
        y = y.to(self.device)
        preds = self._forward(x)
        return {'loss': self.loss(preds, y)}

    def _predict_on_batch(self, x: torch.Tensor, proba: bool) -> List:
        """Returns prediction on batch."""
        assert not self.model.training, "model must be in eval mode"
        x = x.to(self.device)
        pred = self.model(x)
        if proba:
            pred = softmax(pred, dim=1).cpu().detach().tolist()
        else:
            pred = pred.argmax(1).cpu().detach().tolist()
        return pred


class ObjectDetectionExperimenter(NNExperimenter):
    """Class for working with object detection models.

    Args:
        model: Trainable model.
        metric: Target metric by which models are compared.
            One of ``'map'``, ``'map_50'``, ``'map_75'``.
        name: Name of the model.
        weights: Path to the model state_dict to load weights.
        device: String passed to ``torch.device`` initialization.
    """

    def __init__(
            self,
            model: torch.nn.Module,
            metric: str = 'map',
            name: Optional[str] = None,
            weights: Optional[str] = None,
            device: str = 'cuda',
    ):
        parameter_value_check(
            parameter='metric',
            value=metric,
            valid_values={'map', 'map_50', 'map_75'},
        )
        super().__init__(
            model=model,
            metric=metric,
            metric_counter=ObjectDetectionMetricCounter,
            name=name,
            weights=weights,
            device=device
        )

    def _forward(self, x: torch.Tensor) -> List:
        """Implements the forward method of the model and returns predictions."""
        assert not self.model.training
        images = list(image.to(self.device) for image in x)
        preds = self.model(images)
        return [{k: v.to('cpu').detach() for k, v in p.items()} for p in preds]

    def _forward_with_loss(self, x: torch.Tensor, y) -> Dict[str, torch.Tensor]:
        """Implements the train forward method and returns loss."""
        assert self.model.training, "model must be in training mode"
        images = [image.to(self.device) for image in x]
        targets = [{k: v.to(self.device) for k, v in target.items()} for target in y]
        return self.model(images, targets)

    def _predict_on_batch(self, x: torch.Tensor, proba: bool) -> List:
        """Returns prediction on batch."""
        assert not self.model.training, "model must be in eval mode"
        images = [image.to(self.device) for image in x]
        preds = self.model(images)
        if not proba:
            for pred in preds:
                not_thresh = pred['scores'] > 0.5
                pred['boxes'] = pred['boxes'][not_thresh]
                pred['labels'] = pred['labels'][not_thresh]
                pred.pop('scores')
        preds = [{k: v.tolist() for k, v in p.items()} for p in preds]
        return preds


class SegmentationExperimenter(NNExperimenter):
    """Class for working with semantic segmentation models.

    Args:
        model: Trainable model.
        metric: Target metric by which models are compared.
            One of ``'iou'``, ``'dice'``.
        loss: Loss function applied to model output.
        name: Name of the model.
        weights: Path to the model state_dict to load weights.
        device: String passed to ``torch.device`` initialization.
    """

    def __init__(
            self,
            model: torch.nn.Module,
            metric: str = 'iou',
            loss: Callable = torch.nn.CrossEntropyLoss(),
            name: Optional[str] = None,
            weights: Optional[str] = None,
            device: str = 'cuda',
    ):
        parameter_value_check(
            parameter='metric',
            value=metric,
            valid_values={'iou', 'dice'},
        )
        super().__init__(
            model=model,
            metric=metric,
            metric_counter=SegmentationMetricCounter,
            name=name,
            weights=weights,
            device=device
        )
        self.loss = loss

    def _forward(self, x):
        """Implements the forward method of the model and returns predictions."""
        x = x.to(self.device)
        return self.model(x)['out'].to('cpu').detach()

    def _forward_with_loss(self, x, y) -> Dict[str, torch.Tensor]:
        """Implements the train forward method and returns loss."""
        x = x.to(self.device)
        y = y.to(self.device)
        preds = self.model(x)['out']
        return {'loss': self.loss(preds, y)}

    def _predict_on_batch(self, x, proba: bool) -> List:
        """Returns prediction on batch."""
        assert not self.model.training, "model must be in eval mode"
        x = x.to(self.device)
        pred = torch.sigmoid(self.model(x)).cpu().detach()
        if not proba:
            pred = pred.argmax(1)
        return pred



try:
    import seaborn
except:
    pass
import math
from multiprocessing import cpu_count
from typing import TypeVar, Optional

import numpy as np
from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from joblib import Parallel, delayed
from matplotlib.pylab import rcParams
from pymonad.either import Either
from statsforecast.arima import AutoARIMA

from fedot_ind.core.operation.decomposition.spectrum_decomposition import SpectrumDecomposer
from fedot_ind.core.operation.transformation.data.hankel import HankelMatrix
from fedot_ind.core.operation.transformation.regularization.spectrum import sv_to_explained_variance_ratio
from fedot_ind.core.operation.transformation.window_selector import WindowSizeSelector

class_type = TypeVar("T", bound="DataDrivenBasis")
rcParams['figure.figsize'] = 11, 4


class SSAForecasterImplementation(ModelImplementation):
    """
    Model for forecasting uni-variate timeseries with Singular Spectrum Decomposition.
    For given time series T we construct trajectory matrix (hankel matrix) X.
    X = U x S x V_t. After decomposition, we forecast V_t rows separately and after it reconstruct basis. Note that we
    use only few components to reconstruct basis. Other components considered as error (we just sample them).

    Attributes:
        self.window_size_method: str, method for estimating window size for SSA forecaster

    Example:
        To use this operation you can create pipeline as follows::
            import numpy as np
            from fedot.core.data.data import InputData
            from fedot.core.pipelines.pipeline_builder import PipelineBuilder
            from examples.example_utils import get_ts_data
            from fedot_ind.core.repository.initializer_industrial_models import IndustrialModels

            forecast_length = 13
            train_data, test_data, dataset_name = get_ts_data('m4_monthly', forecast_length)
            with IndustrialModels():
                pipeline = PipelineBuilder().add_node('ssa_forecaster').build()
                pipeline.fit(train_data)
                prediction = pipeline.predict(test_data)
                print(prediction)
    """
    LAST_VALUES_THRESHOLD = 100

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)

        self.window_size_method = params.get('window_size_method')
        self.n_processes = math.ceil(cpu_count() * 0.7) if cpu_count() > 1 else 1
        self._SV_threshold = None
        self._decomposer = None
        self._window_size = None

    def predict(self, input_data: InputData) -> OutputData:
        forecast_length = input_data.task.task_params.forecast_length
        features = input_data.features[-self.LAST_VALUES_THRESHOLD:]
        trajectory_transformer = HankelMatrix(time_series=features, window_size=self._window_size)
        data = trajectory_transformer.trajectory_matrix

        self._window_size = trajectory_transformer.window_length
        self._decomposer = SpectrumDecomposer(data,
                                              trajectory_transformer.ts_length + forecast_length)
        U, s, VT = self.get_svd(data)

        s_basis = s[:self._SV_threshold]

        parallel = Parallel(n_jobs=self.n_processes, verbose=0, pre_dispatch="2*n_jobs")
        new_VT = np.array(
            parallel(delayed(self._predict_component)(sample, forecast_length) for sample in VT[:s_basis.shape[0]]))
        basis = self.reconstruct_basis(U, s_basis, new_VT).T

        summed_basis = np.array(basis).sum(axis=0)
        reconstructed_forecast = summed_basis[-forecast_length:]
        reconstructed_features = summed_basis[:-forecast_length]
        error = features[-forecast_length:] - reconstructed_features[-forecast_length:]

        return reconstructed_forecast + error

    def fit(self, input_data: InputData):
        pass

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        features = input_data.features[-self.LAST_VALUES_THRESHOLD:]
        self._window_size = int(WindowSizeSelector(method=self.window_size_method).get_window_size(features) * len(
            features) / 100)
        trajectory_transformer = HankelMatrix(time_series=features, window_size=self._window_size)
        data = trajectory_transformer.trajectory_matrix
        self._decomposer = SpectrumDecomposer(data, trajectory_transformer.ts_length)

        self._window_size = trajectory_transformer.window_length
        self._decomposer = SpectrumDecomposer(data,
                                              trajectory_transformer.ts_length)
        U, s, VT = self.get_svd(data)
        basis = self.reconstruct_basis(U, s, VT).T
        reconstructed_features = np.array(basis).sum(axis=0)
        return reconstructed_features

    def _predict_component(self, comp: np.array, forecast_length: int):
        estimated_seasonal_length = WindowSizeSelector('hac').get_window_size(comp)
        model_arima = AutoARIMA(period=estimated_seasonal_length)
        forecast = []
        for model in [model_arima]:
            model.fit(comp)
            p = model.predict(forecast_length)['mean']
            forecast.append(p)
        forecast = np.mean(np.array(forecast), axis=0)
        return np.concatenate([comp, forecast])

    def get_svd(self, data):
        components = Either.insert(data).then(self._decomposer.svd).then(self._decomposer.threshold).value[0]
        return components

    def reconstruct_basis(self, U, s, VT):
        return Either.insert([U, s, VT]).then(self._decomposer.data_driven_basis).value[0]



import os
from typing import Optional

import numpy as np
from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import \
    DataOperationImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum

from fedot_ind.api.utils.path_lib import PROJECT_PATH
from fedot_ind.core.operation.caching import DataCacher


class IndustrialCachableOperationImplementation(DataOperationImplementation):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)

        cache_folder = os.path.join(PROJECT_PATH, 'cache')
        os.makedirs(cache_folder, exist_ok=True)
        self.cacher = DataCacher(data_type_prefix=f'Features of basis',
                                 cache_folder=cache_folder)

        self.data_type = DataTypesEnum.image

    def fit(self, data):
        """Decomposes the given data on the chosen basis.

        Returns:
            np.array: The decomposition of the given data.
        """
        pass

    def try_load_from_cache(self, hashed_info: str) -> np.array:
        predict = self.cacher.load_data_from_cache(hashed_info=hashed_info)
        return predict

    def transform(self, input_data: InputData, use_cache: bool = False) -> OutputData:
        """Method firstly tries to load result from cache. If unsuccessful, it starts to generate features
        """
        if use_cache:
            class_params = {k: v for k, v in self.__dict__.items() if k not in ['cacher',
                                                                                'data_type',
                                                                                'params',
                                                                                'n_processes',
                                                                                'logging_params',
                                                                                'logger',
                                                                                'relevant_features']}

            hashed_info = self.cacher.hash_info(data=input_data.features,
                                                operation_info=class_params.__repr__())
            try:
                predict = self.try_load_from_cache(hashed_info)
            except FileNotFoundError:
                predict = self._transform(input_data)
                self.cacher.cache_data(hashed_info, predict)

            predict = self._convert_to_output(input_data, predict, data_type=self.data_type)
            return predict
        else:
            predict = self._transform(input_data)
            predict = self._convert_to_output(input_data, predict, data_type=self.data_type)
            return predict

    def _transform(self, input_data):
        pass



from typing import Dict, Optional, Union, List, Tuple

import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch

_PALETTE = ((255, 0, 0), (0, 255, 0), (0, 0, 255), (255, 255, 0), (0, 255, 255), (255, 0, 255), (128, 0, 0),
            (0, 128, 0), (128, 0, 128), (0, 128, 128), (0, 0, 128), (128, 0, 0), (220, 20, 60), (255, 165, 0),
            (218, 165, 32), (240, 230, 140), (154, 205, 50), (107, 142, 35), (0, 100, 0), (46, 139, 87),
            (32, 178, 170), (64, 224, 208), (70, 130, 180), (138, 43, 226), (72, 61, 139), (147, 112, 219),
            (139, 0, 139), (218, 112, 214), (219, 112, 147), (255, 20, 147), (255, 228, 196), (139, 69, 19),
            (210, 105, 30), (244, 164, 96), (188, 143, 143), (112, 128, 144), (230, 230, 250), (245, 245, 245))


def _2nparray(arrs: List[Union[torch.Tensor, List]]) -> List[np.ndarray]:
    return [arr.numpy() if isinstance(arr, torch.Tensor) else np.array(arr) for arr in arrs]


def draw_sample_with_bboxes(
        image: Union[torch.Tensor, str],
        target: Optional[Dict[str, Union[torch.Tensor, List]]] = None,
        prediction: Optional[Dict[str, Union[torch.Tensor, List]]] = None,
        threshold: float = 0.5
) -> plt.Figure:
    """
    Returns the image with bounding boxes.

    Args:
        image: image tensor or path to image.
        target: Dictionary of target values with keys ``'boxes'`` and ``'labels'``.
        prediction: Dictionary of predicted values with keys ``'boxes'``, ``'labels'`` and ``'scores'``.
        threshold: Confidence threshold for displaying predicted bounding boxes.

    Returns:
        `matplotlib.pyplot.Figure` of the image with bounding boxes.
    """
    assert prediction is not None or target is not None, "At least one parameter from 'target' and 'prediction' must not be None"

    if isinstance(image, torch.Tensor):
        image = image.permute(1, 2, 0).numpy()
    else:
        image = plt.imread(image)

    n = 1 if prediction is None or target is None else 2
    fig = plt.figure(figsize=(10 * n, 10))

    thickness = 1 + int(image.shape[-2] / 500)
    font_scale = image.shape[-2] / 1000

    if target is not None:
        ax = plt.subplot(1, n, 1)
        boxes, labels = _2nparray([target['boxes'], target['labels']])
        timage = image.copy()
        for box, label in zip(boxes.astype(np.int32), labels.astype(str)):
            cv2.rectangle(timage, (box[0], box[1]), (box[2], box[3]), (220, 255, 255), thickness)
            cv2.putText(timage, label, (box[0], box[1]), 0, font_scale, (255, 255, 255), thickness)
        ax.set_axis_off()
        ax.imshow(timage)

    if prediction is not None:
        ax = plt.subplot(1, n, n)
        boxes, labels, scores = _2nparray([prediction['boxes'], prediction['labels'], prediction['scores']])

        not_thresh = scores > threshold
        boxes = boxes[not_thresh]
        labels = labels[not_thresh]
        scores = scores[not_thresh]

        pimage = image.copy()
        for box, label, score in zip(boxes.astype(np.int32), labels, scores):
            cv2.rectangle(pimage, (box[0], box[1]), (box[2], box[3]), (220, 255, 255), thickness)
            cv2.putText(pimage, f'{label} ({score:.2f})', (box[0], box[1]), 0, font_scale, (255, 255, 255), thickness)
        ax.set_axis_off()
        ax.imshow(pimage)
    return fig


def _put_mask(axis: plt.axis, image: np.ndarray, mask: np.ndarray, palette: Tuple):
    thickness = 1 + int(image.shape[-2] / 500)
    image = image.copy()
    for ch in range(mask.shape[0]):
        contours, _ = cv2.findContours(mask[ch, :, :], cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)
        for contour in contours:
            cv2.polylines(image, contour, True, palette[ch], thickness)
    axis.set_axis_off()
    axis.imshow(image)


def draw_sample_with_masks(
        image: torch.Tensor,
        target: torch.Tensor = None,
        prediction: torch.Tensor = None,
        palette: Tuple = _PALETTE
) -> plt.Figure:
    """
    Returns the image with masks.

    Args:
        image: Image tensor.
        target: N-channel target tensor with masks, where n - number of classes.
        prediction: N-channel prediction tensor with masks, where n - number of classes.
        palette: Color palette for each class.

    Returns:
        `matplotlib.pyplot.Figure` of the image with masks.
    """
    assert prediction is not None or target is not None, "At least one parameter from 'target' and 'prediction' must not be None"

    image = image.permute(1, 2, 0).numpy()

    n = 1 if prediction is None or target is None else 2
    fig = plt.figure(figsize=(10 * n, 10))

    if target is not None:
        ax = plt.subplot(1, n, 1)
        _put_mask(ax, image, target.numpy().astype(np.uint8), palette)

    if prediction is not None:
        ax = plt.subplot(1, n, n)
        _put_mask(ax, image, prediction.numpy().astype(np.uint8), palette)
    return fig
